# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

from fortls import __version__  # noqa: E402

# Generate the agglomerated changes (from the CHANGELOG) between fortls
# and the fortran-language-server project
with open("../CHANGELOG.md", "r") as f:
    lns = f.readlines()

lns = lns[0 : lns.index("## 1.12.0\n")]
changes = {
    "Added": [],
    "Changed": [],
    "Deprecated": [],
    "Removed": [],
    "Fixed": [],
    "Security": [],
}

field = ""
for i in lns:
    if i.startswith("## "):
        continue
    if i.startswith("### "):
        field = i[4:-1]
        continue
    if i.startswith("- ") or i.startswith("  "):
        changes[field].append(i)

new_file = ["# Unique fortls features (not in fortran-language-server)\n"]
for key, val in changes.items():
    if val:
        new_file.append(f"\n## {key}\n\n")
        new_file.extend(val)

with open("fortls_changes.md", "w") as f:
    f.writelines(new_file)


# -- Project information -----------------------------------------------------

project = "fortls"
copyright = "2021, Giannis Nikiteas"
author = "Giannis Nikiteas"

# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinxarg.ext",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.inheritance_diagram",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosectionlabel",
    "myst_parser",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = [".rst", ".md"]


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"
html_theme = "sphinx_rtd_theme"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


display_toc = True
# autodoc_default_flags = ["members"]
autosummary_generate = True


intersphinx_mapping = {
    "python": ("https://docs.python.org/3.10", None),
}

inheritance_graph_attrs = {
    "size": '"6.0, 8.0"',
    "fontsize": 32,
    "bgcolor": "transparent",
}
inheritance_node_attrs = {
    "color": "black",
    "fillcolor": "white",
    "style": '"filled,solid"',
}
inheritance_edge_attrs = {
    "penwidth": 1.2,
    "arrowsize": 0.8,
}



# from types import NoneType
from setup_tests import path_to_uri, run_request, test_dir, write_rpc_request

from fortls.json_templates import uri_json


def imp_request(file, line, char):
    return write_rpc_request(
        1,
        "textDocument/implementation",
        {
            "textDocument": {"uri": path_to_uri(str(file))},
            "position": {"line": line, "character": char},
        },
    )


def check_imp_request(response: dict, references: dict):
    for uri, changes in response.items():
        refs = references[uri]
        # Loop over all the changes in the current URI, instances of object
        for c, r in zip(changes, refs):
            assert c["range"] == r["range"]


def create(file, line, schar, echar):
    return uri_json(path_to_uri(str(file)), line, schar, line, echar)


def test_implementation_type_bound():
    """Go to implementation of type-bound procedures"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test.f90"
    string += imp_request(file_path, 3, 17)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    assert results[1] == create(test_dir / "subdir" / "test_free.f90", 49, 11, 28)


def test_implementation_intrinsics():
    """Go to implementation of implicit methods is handled gracefully"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_intrinsic.f90"
    string += imp_request(file_path, 11, 18)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    assert results[1] is None


def test_implementation_integer():
    """Go to implementation when no implementation is present is handled gracefully"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_intrinsic.f90"
    string += imp_request(file_path, 20, 31)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    assert results[1] is None


def test_implementation_empty():
    """Go to implementation for empty lines is handled gracefully"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_intrinsic.f90"
    string += imp_request(file_path, 13, 0)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    assert results[1] is None


def test_implementation_no_file():
    """Go to implementation for empty lines is handled gracefully"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "fake.f90"
    string += imp_request(file_path, 13, 0)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    assert results[1] is None



from setup_tests import run_request, test_dir, write_rpc_request


def check_return(result_array, checks):
    comm_lines = []
    for (i, hover_line) in enumerate(result_array["contents"][0]["value"].splitlines()):
        if hover_line.count("!!") > 0:
            comm_lines.append((i, hover_line))
    assert len(comm_lines) == len(checks)
    for i in range(len(checks)):
        assert comm_lines[i][0] == checks[i][0]
        assert comm_lines[i][1] == checks[i][1]


def hover_request(file_path, line, char):
    return write_rpc_request(
        1,
        "textDocument/hover",
        {
            "textDocument": {"uri": str(file_path)},
            "position": {"line": line, "character": char},
        },
    )


def test_doxygen():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "docs")})
    file_path = test_dir / "docs" / "test_doxygen.f90"
    string += hover_request(file_path, 15, 17)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = (
        (1, "!! @brief inserts a value into an ordered array"),
        (2, "!! "),
        (
            3,
            '!! An array "list" consisting of n ascending ordered values. The method'
            " insert a",
        ),
        (4, '!! "new_entry" into the array.'),
        (5, "!! hint: use cshift and eo-shift"),
        (6, "!! "),
        (7, "!! @param[in,out]   list    a real array, size: max_size"),
        (8, "!! @param[in]       n       current values in the array"),
        (9, "!! @param[in]       max_size    size if the array"),
        (10, "!! @param[in]       new_entry   the value to insert"),
    )
    check_return(results[1], ref)


def test_ford():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "docs")})
    file_path = test_dir / "docs" / "test_ford.f90"
    string += hover_request(file_path, 5, 20)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = (
        (1, "!! Feeds your cats and dogs, if enough food is available. If not enough"),
        (2, "!! food is available, some of your pets will get angry."),
        (4, " !! The number of cats to keep track of."),
        (6, " !! The number of dogs to keep track of."),
        (8, " !! The amount of pet food (in kilograms) which you have on hand."),
        (10, " !! The number of pets angry because they weren't fed."),
    )
    check_return(results[1], ref)


def test_doc_overwrite_type_bound_procedure_sub():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # Test we can override method docstring e.g.
    # procedure :: name => name_imp !< Doc override
    # We want to preserve the argument list docstring
    string += hover_request(file_path, 13, 19)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((1, "!! Doc 1"), (3, " !! Doc 5")))


def test_doc_type_bound_procedure_sub_implementation():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # procedure :: name => name_imp !< Doc override
    # Test that name_imp will yield the full docstring present in the implementation
    string += hover_request(file_path, 13, 31)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((1, "!! Doc 4"), (4, " !! Doc 5")))


def test_doc_overwrite_type_bound_procedure_fun():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # Test we can override function docstring e.g.
    # procedure :: name => name_imp !< Doc override
    # We want to preserve the argument list docstring
    string += hover_request(file_path, 14, 17)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((1, "!! Doc 2"),))


def test_doc_type_bound_procedure_fun_implementation():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # procedure :: name => name_imp !< Doc override
    # Test that name_imp will yield the full docstring present in the implementation
    string += hover_request(file_path, 14, 28)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((1, "!! Doc 6"),))


def test_doc_empty_overwrite_type_bound_procedure_sub():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # Test we can ignore overriding method docstring  and return the original e.g.
    # procedure :: name => name_imp !<
    # We want to preserve the argument list docstring
    string += hover_request(file_path, 21, 18)
    string += hover_request(file_path, 21, 37)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((1, "!! Doc 7"), (3, " !! Doc 8")))
    check_return(results[2], ((1, "!! Doc 7"), (4, " !! Doc 8")))


def test_doc_empty_overwrite_type_bound_procedure_fun():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # procedure :: name => name_imp !< Doc
    # We want to preserve the procedure docstring but also fetch the empty
    # docs for the implementation
    string += hover_request(file_path, 22, 17)
    string += hover_request(file_path, 22, 32)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((1, "!! Doc 3"),))
    check_return(results[2], ())


def test_doc_multiline_type_bound_procedure_arg_list():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    # Check that inline docstrings can be input and carried over in multiple lines
    # for both the procedure pointer and the implementation
    string += hover_request(file_path, 15, 32)
    string += hover_request(file_path, 15, 47)
    errcode, results = run_request(string)
    assert errcode == 0
    check_return(results[1], ((2, " !! Doc 9"), (3, " !! Doc 10")))
    check_return(
        results[2],
        ((2, " !! Doc 9"), (3, " !! Doc 10"), (5, " !! Doc 11"), (6, " !! Doc 12")),
    )



from setup_tests import (
    check_post_msg,
    path_to_uri,
    run_request,
    test_dir,
    write_rpc_request,
)


def rename_request(new_name: str, file_path, ln: int, ch: int):
    return write_rpc_request(
        1,
        "textDocument/rename",
        {
            "newName": new_name,
            "textDocument": {"uri": str(file_path)},
            "position": {"line": ln, "character": ch},
        },
    )


def check_rename_response(response: dict, references: dict):
    # Loop over URI's if the change spans multiple files there will be more than 1
    for uri, changes in response.items():
        refs = references[uri]
        # Loop over all the changes in the current URI, instances of object
        for c, r in zip(changes, refs):
            assert c["range"] == r["range"]
            assert c["newText"] == r["newText"]


def create(new_text: str, sln: int, sch: int, eln: int, ech: int):
    return {
        "range": {
            "start": {"line": sln, "character": sch},
            "end": {"line": eln, "character": ech},
        },
        "newText": new_text,
    }


def test_rename_var():
    """ "Test simple variable rename"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += rename_request("str_rename", file_path, 5, 25)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = {}
    ref[path_to_uri(str(file_path))] = [create("str_rename", 5, 20, 5, 29)]
    check_rename_response(results[1]["changes"], ref)


def test_rename_var_across_module():
    """Test renaming objects like variables across modules works"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += rename_request("new_module_var", file_path, 26, 15)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = {}
    ref[path_to_uri(str(test_dir / "subdir" / "test_free.f90"))] = [
        create("new_module_var", 32, 11, 32, 26)
    ]
    ref[path_to_uri(str(file_path))] = [create("new_module_var", 2, 44, 2, 59)]
    ref[path_to_uri(str(file_path))].append(create("new_module_var", 26, 8, 26, 23))

    check_rename_response(results[1]["changes"], ref)


def test_rename_empty():
    """Test that renaming nothing will not error"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_imp_type_bound_proc.f90"
    string += rename_request("bar", file_path, 9, 0)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    assert results[1] is None


def test_rename_member_type_ptr():
    """Test that renaming type bound pointers of procedure methods rename
    only the pointer and not the implementation, even if the pointer and the
    implementation share the same name
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += rename_request("bp_rename", file_path, 18, 25)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = {}
    ref[path_to_uri(str(file_path))] = [create("bp_rename", 18, 16, 18, 26)]
    ref[path_to_uri(str(test_dir / "subdir" / "test_free.f90"))] = [
        create("bp_rename", 15, 27, 15, 37)
    ]
    check_rename_response(results[1]["changes"], ref)


def test_rename_member_type_ptr_null():
    """Test renaming type bound pointers of procedure methods works when pointing to
    null
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += rename_request("bp_rename", file_path, 17, 25)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = {}
    ref[path_to_uri(str(file_path))] = [create("bp_rename", 17, 16, 17, 28)]
    ref[path_to_uri(str(test_dir / "subdir" / "test_free.f90"))] = [
        create("bp_rename", 11, 43, 11, 55)
    ]
    check_rename_response(results[1]["changes"], ref)


def test_rename_type_bound_proc_no_ptr():
    """Test renaming type bound pointers of procedure methods works when no pointer
    is setup. Requesting to rename the procedure should rename, the implementation
    and the Method itself i.e. call self%foo()

    Requesting to rename the implementation should also rename the procedure
    and all the locations it is called in
    """
    root = test_dir / "rename"
    string = write_rpc_request(1, "initialize", {"rootPath": str(root)})
    file_path = root / "test_rename_imp_type_bound_proc.f90"
    # Rename the procedure name and check if implementation also renames
    string += rename_request("bar", file_path, 5, 23)
    # Rename the implementation name and check if declaration, references change
    string += rename_request("bar", file_path, 10, 18)
    errcode, results = run_request(string)
    assert errcode == 0
    ref = {}
    ref[path_to_uri(str(file_path))] = [create("bar", 5, 21, 5, 24)]
    ref[path_to_uri(str(file_path))].append(create("bar", 10, 15, 10, 18))
    ref[path_to_uri(str(file_path))].append(create("bar", 12, 18, 12, 21))
    ref[path_to_uri(str(file_path))].append(create("bar", 13, 19, 13, 22))
    check_rename_response(results[1]["changes"], ref)
    check_rename_response(results[2]["changes"], ref)


def test_rename_non_existent_file():
    """Test renaming type bound pointers of procedure methods works when pointing to
    null
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "fake.f90"
    string += rename_request("bar", file_path, 5, 23)
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1] is None


def test_rename_nested():
    """Test renaming heavily nested constructs works"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_nested.f90"
    string += rename_request("bar", file_path, 6, 23)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    ref = {}
    ref[path_to_uri(str(file_path))] = [create("bar", 6, 23, 6, 26)]
    ref[path_to_uri(str(file_path))].append(create("bar", 9, 27, 9, 30))
    check_rename_response(results[1]["changes"], ref)


def test_rename_intrinsic():
    """Test renaming an intrinsic function, while no other function exists
    with the same name, will throw an error
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_nested.f90"
    string += rename_request("bar", file_path, 8, 27)
    errcode, results = run_request(string, ["-n", "1"])
    assert errcode == 0
    check_post_msg(results[1], "Rename failed: Cannot rename intrinsics", 2)
    assert results[2] is None


def test_rename_use_only_rename():
    """Test renaming constructs of `use mod, only: val => root_val
    are handled correctly
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "subdir")})
    file_path = test_dir / "subdir" / "test_rename.F90"
    string += rename_request("bar", file_path, 13, 5)
    errcode, results = run_request(string, ["-n", "1"])
    # FIXME: to be implemented
    assert errcode == 0


def test_rename_skip_intrinsic():
    """Test that renaming functions named the same as intrinsic functions e.g. size()
    will only rename the user defined functions
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "rename")})
    file_path = test_dir / "rename" / "test_rename_intrinsic.f90"
    string += rename_request("bar", file_path, 22, 13)
    errcode, results = run_request(string, ["-n", "1"])
    # FIXME: to be implemented
    assert errcode == 0



from setup_tests import run_request, test_dir, write_rpc_notification, write_rpc_request


def test_init():
    def check_return(result_dict):
        # Expected capabilities
        # {
        #     "completionProvider": {
        #         "resolveProvider": false,
        #         "triggerCharacters": ["%"]
        #     },
        #     "definitionProvider": true,
        #     "documentSymbolProvider": true,
        #     "referencesProvider": True,
        #     "hoverProvider": true,
        #     "textDocumentSync": 2
        # }
        #
        assert "capabilities" in result_dict
        assert result_dict["capabilities"]["textDocumentSync"] == 2
        assert result_dict["capabilities"]["definitionProvider"] is True
        assert result_dict["capabilities"]["documentSymbolProvider"] is True
        assert result_dict["capabilities"]["hoverProvider"] is True
        assert result_dict["capabilities"]["referencesProvider"] is True
        assert (
            result_dict["capabilities"]["completionProvider"]["resolveProvider"]
            is False
        )
        assert (
            result_dict["capabilities"]["completionProvider"]["triggerCharacters"][0]
            == "%"
        )

    #
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    errcode, results = run_request(string)
    #
    assert errcode == 0
    check_return(results[0])


def test_logger():
    """Test the logger"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    errcode, results = run_request(string, ["--debug_log"])
    assert errcode == 0
    assert results[1]["type"] == 3
    assert results[1]["message"] == "fortls debugging enabled"


def test_open():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = str(test_dir / "subdir" / "test_free.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string, fortls_args=["--disable_diagnostics"])
    #
    assert errcode == 0
    assert len(results) == 1


def test_change():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_unknown.f90"
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": str(file_path)}}
    )
    string += write_rpc_notification(
        "textDocument/didChange",
        {
            "textDocument": {"uri": str(file_path)},
            "contentChanges": [
                {
                    "text": "module test_unkown\nend module test_unknown\n",
                    "range": {
                        "start": {"line": 0, "character": 0},
                        "end": {"line": 0, "character": 0},
                    },
                }
            ],
        },
    )
    string += write_rpc_request(
        2, "textDocument/documentSymbol", {"textDocument": {"uri": str(file_path)}}
    )
    file_path = test_dir / "subdir" / "test_free.f90"
    string += write_rpc_notification(
        "textDocument/didChange",
        {
            "textDocument": {"uri": str(file_path)},
            "contentChanges": [
                {
                    "text": " unicode test",
                    "range": {
                        "start": {"line": 3, "character": 3},
                        "end": {"line": 3, "character": 3},
                    },
                },
                {
                    "text": "",
                    "range": {
                        "start": {"line": 6, "character": 0},
                        "end": {"line": 31, "character": 0},
                    },
                },
                {
                    "text": "",
                    "range": {
                        "start": {"line": 7, "character": 0},
                        "end": {"line": 39, "character": 0},
                    },
                },
            ],
        },
    )
    string += write_rpc_request(
        3, "textDocument/documentSymbol", {"textDocument": {"uri": str(file_path)}}
    )
    errcode, results = run_request(string, fortls_args=["--disable_diagnostics"])
    #
    assert errcode == 0
    assert len(results) == 3
    assert len(results[1]) == 1
    assert len(results[2]) == 5


def test_symbols():
    def check_return(result_array):
        # Expected objects
        objs = (
            ["test_free", 2, 0, 81],
            ["scale_type", 5, 4, 6],
            ["val", 13, 5, 5],
            ["vector", 5, 8, 16],
            ["n", 13, 9, 9],
            ["v", 13, 10, 10],
            ["bound_nopass", 6, 11, 11],
            ["create", 6, 13, 13],
            ["norm", 6, 14, 14],
            ["bound_pass", 6, 15, 15],
            ["scaled_vector", 5, 18, 23],
            ["scale", 13, 19, 19],
            ["set_scale", 6, 21, 21],
            ["norm", 6, 22, 22],
            ["fort_wrap", 11, 26, 29],
            ["vector_create", 12, 35, 41],
            ["vector_norm", 12, 43, 47],
            ["scaled_vector_set", 12, 49, 53],
            ["scaled_vector_norm", 12, 55, 59],
            ["unscaled_norm", 12, 61, 65],
            ["test_sig_Sub", 12, 67, 70],
            ["bound_pass", 12, 72, 80],
        )
        assert len(result_array) == len(objs)
        for i, obj in enumerate(objs):
            assert result_array[i]["name"] == obj[0]
            assert result_array[i]["kind"] == obj[1]
            assert result_array[i]["location"]["range"]["start"]["line"] == obj[2]
            assert result_array[i]["location"]["range"]["end"]["line"] == obj[3]

    #
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    string += write_rpc_request(
        2, "textDocument/documentSymbol", {"textDocument": {"uri": str(file_path)}}
    )
    errcode, results = run_request(string)
    #
    assert errcode == 0
    check_return(results[1])


def test_workspace_symbols():
    def check_return(result_array):
        # Expected objects
        objs = (
            ["test", 6, 7],
            ["test_abstract", 2, 0],
            ["test_free", 2, 0],
            ["test_gen_type", 5, 1],
            ["test_generic", 2, 0],
            ["test_inherit", 2, 0],
            ["test_int", 2, 0],
            ["test_mod", 2, 0],
            ["test_nan", 2, 0],
            ["test_nonint_mod", 2, 0],
            ["test_preproc_keywords", 2, 0],
            ["test_private", 2, 8],
            ["test_program", 2, 0],
            ["test_rename_sub", 6, 9],
            ["test_select", 2, 0],
            ["test_select_sub", 6, 16],
            ["test_sig_Sub", 6, 67],
            ["test_str1", 13, 5],
            ["test_str2", 13, 5],
            ["test_sub", 6, 8],
            ["test_vis_mod", 2, 0],
        )
        assert len(result_array) == len(objs)
        for i, obj in enumerate(objs):
            assert result_array[i]["name"] == obj[0]
            assert result_array[i]["kind"] == obj[1]
            assert result_array[i]["location"]["range"]["start"]["line"] == obj[2]

    #
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    string += write_rpc_request(2, "workspace/symbol", {"query": "test"})
    errcode, results = run_request(string)
    #
    assert errcode == 0
    check_return(results[1])



# TODO: enable jsonc C-style comments
from __future__ import annotations

import json
import logging
import os
import re
import subprocess
import sys
import traceback
import urllib.request
from multiprocessing import Pool
from pathlib import Path
from typing import Pattern
from urllib.error import URLError

from packaging import version

# Local modules
from fortls.constants import (
    CLASS_TYPE_ID,
    FORTRAN_LITERAL,
    FUNCTION_TYPE_ID,
    INTERFACE_TYPE_ID,
    METH_TYPE_ID,
    MODULE_TYPE_ID,
    SELECT_TYPE_ID,
    SUBROUTINE_TYPE_ID,
    VAR_TYPE_ID,
    FRegex,
    Severity,
    log,
)
from fortls.helper_functions import (
    expand_name,
    get_line_prefix,
    get_paren_level,
    get_var_stack,
    only_dirs,
    resolve_globs,
    set_keyword_ordering,
)
from fortls.intrinsics import (
    fortran_intrinsic_obj,
    get_intrinsic_keywords,
    load_intrinsics,
    set_lowercase_intrinsics,
)
from fortls.json_templates import change_json, symbol_json, uri_json
from fortls.jsonrpc import JSONRPC2Connection, path_from_uri, path_to_uri
from fortls.objects import (
    climb_type_tree,
    find_in_scope,
    find_in_workspace,
    fortran_ast,
    fortran_var,
    get_use_tree,
)
from fortls.parse_fortran import fortran_file, get_line_context
from fortls.regex_patterns import src_file_exts
from fortls.version import __version__

# Global regexes
# TODO: I think this can be replaced by fortls.regex_patterns type & class
TYPE_DEF_REGEX = re.compile(r"[ ]*(TYPE|CLASS)[ ]*\([a-z0-9_ ]*$", re.I)


class LangServer:
    def __init__(self, conn, settings: dict):
        self.conn: JSONRPC2Connection = conn
        self.running: bool = True
        self.root_path: str = None
        self.workspace: dict[str, fortran_file] = {}
        self.obj_tree: dict = {}
        self.link_version = 0
        self._version = version.parse(__version__)
        # Parse a dictionary of the command line interface and make them into
        # class variable. This way the command line and the file interfaces
        # are always on sync, with the same default arguments
        for k, v in settings.items():
            # Do not parse command line debug arguments
            if k.startswith("debug_") and k != "debug_log":
                continue
            setattr(self, k, v)

        self.sync_type: int = 2 if self.incremental_sync else 1
        self.post_messages = []
        self.FORTRAN_SRC_EXT_REGEX: Pattern[str] = src_file_exts()
        # Intrinsic (re-loaded during initialize)
        (
            self.statements,
            self.keywords,
            self.intrinsic_funs,
            self.intrinsic_mods,
        ) = load_intrinsics()
        # Set object settings
        set_keyword_ordering(self.sort_keywords)

    def post_message(self, msg: str, severity: int = Severity.error, exc_info=False):
        self.conn.send_notification(
            "window/showMessage", {"type": severity, "message": msg}
        )
        if severity == 1:
            log.error(msg, exc_info=exc_info)
        elif severity == 2:
            log.warning(msg, exc_info=exc_info)
        elif severity == 3:
            log.info(msg, exc_info=exc_info)

    def run(self):
        # Run server
        while self.running:
            try:
                request = self.conn.read_message()
                self.handle(request)
            except EOFError:
                break
            except Exception as e:
                self.post_message(f"Unexpected error: {e}", exc_info=True)
                break
            else:
                for message in self.post_messages:
                    self.post_message(message[1], message[0])
                self.post_messages = []

    def handle(self, request: dict):
        def noop(request: dict):
            return None

        # Request handler
        log.debug("REQUEST %s %s", request.get("id"), request.get("method"))
        handler = {
            "initialize": self.serve_initialize,
            "textDocument/documentSymbol": self.serve_document_symbols,
            "textDocument/completion": self.serve_autocomplete,
            "textDocument/signatureHelp": self.serve_signature,
            "textDocument/definition": self.serve_definition,
            "textDocument/references": self.serve_references,
            "textDocument/hover": self.serve_hover,
            "textDocument/implementation": self.serve_implementation,
            "textDocument/rename": self.serve_rename,
            "textDocument/didOpen": self.serve_onOpen,
            "textDocument/didSave": self.serve_onSave,
            "textDocument/didClose": self.serve_onClose,
            "textDocument/didChange": self.serve_onChange,
            "textDocument/codeAction": self.serve_codeActions,
            "initialized": noop,
            "workspace/didChangeWatchedFiles": noop,
            "workspace/symbol": self.serve_workspace_symbol,
            "$/cancelRequest": noop,
            "$/setTrace": noop,
            "shutdown": noop,
            "exit": self.serve_exit,
        }.get(request["method"], self.serve_default)

        # We handle notifications differently since we can't respond
        if "id" not in request:
            try:
                handler(request)
            except:
                log.exception("error handling request: %s", request, exc_info=True)
            return
        #
        try:
            resp = handler(request)
        except JSONRPC2Error as e:
            self.conn.write_error(
                request["id"], code=e.code, message=e.message, data=e.data
            )
            log.warning("RPC error handling request %s", request, exc_info=True)
        except Exception as e:
            self.conn.write_error(
                request["id"],
                code=-32603,
                message=str(e),
                data={
                    "traceback": traceback.format_exc(),
                },
            )
            log.warning("error handling request %s", request, exc_info=True)
        else:
            self.conn.write_response(request["id"], resp)

    def serve_initialize(self, request: dict):
        # Setup language server
        params: dict = request["params"]
        self.root_path = path_from_uri(
            params.get("rootUri") or params.get("rootPath") or ""
        )
        self.source_dirs.add(self.root_path)

        self._load_config_file()
        self._config_logger(request)
        self._load_intrinsics()
        self._add_source_dirs()
        if self._update_version_pypi():
            self.post_message(
                "Please restart the server for the new version to activate",
                Severity.info,
            )

        # Initialize workspace
        self.workspace_init()
        log.info(f"fortls - Fortran Language Server {__version__} Initialized")
        #
        server_capabilities = {
            "completionProvider": {
                "resolveProvider": False,
                "triggerCharacters": ["%"],
            },
            "definitionProvider": True,
            "documentSymbolProvider": True,
            "referencesProvider": True,
            "hoverProvider": True,
            "implementationProvider": True,
            "renameProvider": True,
            "workspaceSymbolProvider": True,
            "textDocumentSync": self.sync_type,
        }
        if self.use_signature_help:
            server_capabilities["signatureHelpProvider"] = {
                "triggerCharacters": ["(", ","]
            }
        if self.enable_code_actions:
            server_capabilities["codeActionProvider"] = True
        if self.notify_init:
            self.post_message("fortls initialization complete", Severity.info)
        return {"capabilities": server_capabilities}

    def serve_workspace_symbol(self, request):
        def map_types(type):
            if type == 1:
                return 2
            elif type == 2:
                return 6
            elif type == 3:
                return 12
            elif type == 4:
                return 5
            elif type == 5:
                return 11
            elif type == 6:
                return 13
            elif type == 7:
                return 6
            else:
                return 1

        matching_symbols = []
        query = request["params"]["query"].lower()
        for candidate in find_in_workspace(self.obj_tree, query):
            tmp_out = {
                "name": candidate.name,
                "kind": map_types(candidate.get_type()),
                "location": {
                    "uri": path_to_uri(candidate.file_ast.path),
                    "range": {
                        "start": {"line": candidate.sline - 1, "character": 0},
                        "end": {"line": candidate.eline - 1, "character": 0},
                    },
                },
            }
            # Set containing scope
            if candidate.FQSN.find("::") > 0:
                tmp_list = candidate.FQSN.split("::")
                tmp_out["containerName"] = tmp_list[0]
            matching_symbols.append(tmp_out)
        return sorted(matching_symbols, key=lambda k: k["name"])

    def serve_document_symbols(self, request: dict):
        def map_types(type, in_class: bool = False):
            if type == 1:
                return 2
            elif type in (2, 3):
                if in_class:
                    return 6
                else:
                    return 12
            elif type == 4:
                return 5
            elif type == 5:
                return 11
            elif type == 6:
                return 13
            elif type == 7:
                return 6
            else:
                return 1

        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        path: str = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return []
        # Add scopes to outline view
        test_output = []
        for scope in file_obj.ast.get_scopes():
            if (scope.name[0] == "#") or (scope.get_type() == SELECT_TYPE_ID):
                continue
            scope_tree = scope.FQSN.split("::")
            if len(scope_tree) > 2:
                if scope_tree[1].startswith("#gen_int"):
                    scope_type = 11
                else:
                    continue
            else:
                scope_type = map_types(scope.get_type())

            # Set containing scope
            if scope.FQSN.find("::") > 0:
                tmp_list = scope.FQSN.split("::")
                test_output.append(
                    symbol_json(
                        scope.name,
                        scope_type,
                        uri,
                        scope.sline - 1,
                        0,
                        scope.eline - 1,
                        0,
                        tmp_list[0],
                    )
                )
            else:
                test_output.append(
                    symbol_json(
                        scope.name,
                        scope_type,
                        uri,
                        scope.sline - 1,
                        0,
                        scope.eline - 1,
                        0,
                    )
                )
            # If class add members
            if scope.get_type() == CLASS_TYPE_ID and not self.symbol_skip_mem:
                for child in scope.children:
                    test_output.append(
                        symbol_json(
                            child.name,
                            map_types(child.get_type(), True),
                            uri,
                            child.sline - 1,
                            0,
                            container_name=scope.name,
                        )
                    )
        return test_output

    def serve_autocomplete(self, request: dict):
        #
        def map_types(type: int):
            if type == 1:
                return 9
            elif type == 2:
                return 3
            elif type == 4:
                return 7
            elif type == 6:
                return 6
            else:
                return type

        def set_type_mask(def_value):
            return [def_value if i < 8 else True for i in range(16)]

        def get_candidates(
            scope_list: list,
            var_prefix: str,
            inc_globals: bool = True,
            public_only: bool = False,
            abstract_only: bool = False,
            no_use: bool = False,
        ):
            #
            def child_candidates(
                scope, only_list: list = None, filter_public=True, req_abstract=False
            ):
                if only_list is None:
                    only_list = []
                tmp_list = []
                # Filter children
                nonly = len(only_list)
                for child in scope.get_children(filter_public):
                    if req_abstract:
                        if child.is_abstract():
                            tmp_list += child_candidates(
                                child, only_list, filter_public
                            )
                    else:
                        if child.is_external_int():
                            tmp_list += child_candidates(
                                child, only_list, filter_public
                            )
                        else:
                            if (nonly > 0) and (child.name.lower() not in only_list):
                                continue
                            tmp_list.append(child)
                return tmp_list

            var_list = []
            use_dict = {}
            for scope in scope_list:
                var_list += child_candidates(
                    scope, filter_public=public_only, req_abstract=abstract_only
                )
                # Traverse USE tree and add to list
                if not no_use:
                    use_dict = get_use_tree(scope, use_dict, self.obj_tree)
            # Look in found use modules
            rename_list = [None for _ in var_list]
            for use_mod, use_info in use_dict.items():
                scope = self.obj_tree[use_mod][0]
                only_list = use_info.only_list
                if len(use_info.rename_map) > 0:
                    only_list = [
                        use_info.rename_map.get(only_name, only_name)
                        for only_name in only_list
                    ]
                tmp_list = child_candidates(
                    scope, only_list, req_abstract=abstract_only
                )
                # Setup renaming
                if len(use_info.rename_map) > 0:
                    rename_reversed = {
                        value: key for (key, value) in use_info.rename_map.items()
                    }
                    for tmp_obj in tmp_list:
                        var_list.append(tmp_obj)
                        rename_list.append(
                            rename_reversed.get(tmp_obj.name.lower(), None)
                        )
                else:
                    var_list += tmp_list
                    rename_list += [None for _ in tmp_list]
            # Add globals
            if inc_globals:
                tmp_list = [obj[0] for (_, obj) in self.obj_tree.items()]
                var_list += tmp_list + self.intrinsic_funs
                rename_list += [None for _ in tmp_list + self.intrinsic_funs]
            # Filter by prefix if necessary
            if var_prefix == "":
                return var_list, rename_list
            else:
                tmp_list = []
                tmp_rename = []
                for (i, var) in enumerate(var_list):
                    var_name = rename_list[i]
                    if var_name is None:
                        var_name = var.name
                    if var_name.lower().startswith(var_prefix):
                        tmp_list.append(var)
                        tmp_rename.append(rename_list[i])
                return tmp_list, tmp_rename

        def build_comp(
            candidate,
            name_only: bool = self.autocomplete_name_only,
            name_replace: str = None,
            is_interface: bool = False,
            is_member: bool = False,
        ):
            comp_obj = {}
            call_sig = None
            if name_only:
                comp_obj["label"] = candidate.name
            else:
                comp_obj["label"] = candidate.name
                if name_replace is not None:
                    comp_obj["label"] = name_replace
                call_sig, snippet = candidate.get_snippet(name_replace)
                if self.autocomplete_no_snippets:
                    snippet = call_sig
                if snippet is not None:
                    if self.use_signature_help and (not is_interface):
                        arg_open = snippet.find("(")
                        if arg_open > 0:
                            snippet = snippet[:arg_open]
                    comp_obj["insertText"] = snippet
                    comp_obj["insertTextFormat"] = 2
            comp_obj["kind"] = map_types(candidate.get_type())
            if is_member and (comp_obj["kind"] == 3):
                comp_obj["kind"] = 2
            comp_obj["detail"] = candidate.get_desc()
            if call_sig is not None:
                comp_obj["detail"] += " " + call_sig
            doc_str, _ = candidate.get_hover()
            if doc_str is not None:
                comp_obj["documentation"] = doc_str
            return comp_obj

        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        path: str = path_from_uri(uri)
        file_obj: fortran_file = self.workspace.get(path)
        if file_obj is None:
            return None
        # Check line
        ac_line: int = params["position"]["line"]
        ac_char: int = params["position"]["character"]
        # Get full line (and possible continuations) from file
        pre_lines, curr_line, _ = file_obj.get_code_line(
            ac_line, forward=False, strip_comment=True
        )
        line_prefix = get_line_prefix(pre_lines, curr_line, ac_char)
        if line_prefix is None:
            return None
        is_member = False
        try:
            var_stack = get_var_stack(line_prefix)
            is_member = len(var_stack) > 1
            var_prefix = var_stack[-1].strip()
        except (TypeError, AttributeError):
            return None
        # print(var_stack)
        item_list = []
        # Get context
        name_only = self.autocomplete_name_only
        public_only = False
        include_globals = True
        line_context, context_info = get_line_context(line_prefix)
        if (line_context == "skip") or (var_prefix == "" and (not is_member)):
            return None
        if self.autocomplete_no_prefix:
            var_prefix = ""
        # Suggestions for user-defined type members
        scope_list = []
        if is_member:
            curr_scope = file_obj.ast.get_inner_scope(ac_line + 1)
            type_scope = climb_type_tree(var_stack, curr_scope, self.obj_tree)
            # Set enclosing type as scope
            if type_scope is None:
                return None
            else:
                include_globals = False
                scope_list = [type_scope]
        else:
            scope_list = file_obj.ast.get_scopes(ac_line + 1)
        # Setup based on context
        req_callable = False
        abstract_only = False
        no_use = False
        type_mask = set_type_mask(False)
        type_mask[MODULE_TYPE_ID] = True
        type_mask[CLASS_TYPE_ID] = True
        if line_context == "mod_only":
            # Module names only (USE statement)
            for key in self.obj_tree:
                candidate = self.obj_tree[key][0]
                if (
                    candidate.get_type() == MODULE_TYPE_ID
                ) and candidate.name.lower().startswith(var_prefix):
                    item_list.append(build_comp(candidate, name_only=True))
            return item_list
        elif line_context == "mod_mems":
            # Public module members only (USE ONLY statement)
            name_only = True
            mod_name = context_info.lower()
            if mod_name in self.obj_tree:
                scope_list = [self.obj_tree[mod_name][0]]
                public_only = True
                include_globals = False
                type_mask[CLASS_TYPE_ID] = False
            else:
                return None
        elif line_context == "pro_link":
            # Link to local subroutine/functions
            type_mask = set_type_mask(True)
            type_mask[SUBROUTINE_TYPE_ID] = False
            type_mask[FUNCTION_TYPE_ID] = False
            name_only = True
            include_globals = False
            no_use = True
        elif line_context == "call":
            # Callable objects only ("CALL" statements)
            req_callable = True
        elif line_context == "type_only":
            # User-defined types only (variable definitions, select clauses)
            type_mask = set_type_mask(True)
            type_mask[CLASS_TYPE_ID] = False
        elif line_context == "import":
            # Import statement (variables and user-defined types only)
            name_only = True
            type_mask = set_type_mask(True)
            type_mask[CLASS_TYPE_ID] = False
            type_mask[VAR_TYPE_ID] = False
        elif line_context == "vis":
            # Visibility statement (local objects only)
            include_globals = False
            name_only = True
            type_mask = set_type_mask(True)
            type_mask[CLASS_TYPE_ID] = False
            type_mask[VAR_TYPE_ID] = False
            type_mask[SUBROUTINE_TYPE_ID] = False
            type_mask[FUNCTION_TYPE_ID] = False
            curr_scope = [file_obj.ast.get_inner_scope(ac_line + 1)]
        elif line_context == "int_only":
            # Interfaces only (procedure definitions)
            abstract_only = True
            include_globals = False
            name_only = True
            type_mask = set_type_mask(True)
            type_mask[SUBROUTINE_TYPE_ID] = False
            type_mask[FUNCTION_TYPE_ID] = False
        elif line_context == "var_only":
            # Variables only (variable definitions)
            name_only = True
            type_mask[SUBROUTINE_TYPE_ID] = True
            type_mask[FUNCTION_TYPE_ID] = True
        elif line_context == "var_key":
            # Variable definition keywords only (variable definition)
            key_context = 0
            enc_scope_type = scope_list[-1].get_type()
            if enc_scope_type == MODULE_TYPE_ID:
                key_context = 1
            elif (enc_scope_type == SUBROUTINE_TYPE_ID) or (
                enc_scope_type == FUNCTION_TYPE_ID
            ):
                key_context = 2
            elif enc_scope_type == CLASS_TYPE_ID:
                key_context = 3
            for candidate in get_intrinsic_keywords(
                self.statements, self.keywords, key_context
            ):
                if candidate.name.lower().startswith(var_prefix):
                    item_list.append(build_comp(candidate))
            return item_list
        elif line_context == "first":
            # First word -> default context plus Fortran statements
            for candidate in get_intrinsic_keywords(self.statements, self.keywords, 0):
                if candidate.name.lower().startswith(var_prefix):
                    item_list.append(build_comp(candidate))
        # Build completion list
        candidate_list, rename_list = get_candidates(
            scope_list, var_prefix, include_globals, public_only, abstract_only, no_use
        )
        for (i, candidate) in enumerate(candidate_list):
            # Skip module names (only valid in USE)
            candidate_type = candidate.get_type()
            if type_mask[candidate_type]:
                continue
            if req_callable and (not candidate.is_callable()):
                continue
            #
            name_replace = rename_list[i]
            if candidate_type == INTERFACE_TYPE_ID:
                tmp_list = []
                if name_replace is None:
                    name_replace = candidate.name
                for member in candidate.mems:
                    tmp_text, _ = member.get_snippet(name_replace)
                    if tmp_list.count(tmp_text) > 0:
                        continue
                    tmp_list.append(tmp_text)
                    item_list.append(
                        build_comp(
                            member,
                            name_replace=name_replace,
                            is_interface=True,
                            is_member=is_member,
                        )
                    )
                continue
            #
            item_list.append(
                build_comp(candidate, name_only=name_only, name_replace=name_replace)
            )
        return item_list

    def get_definition(
        self,
        def_file: fortran_file,
        def_line: int,
        def_char: int,
        hover_req: bool = False,
    ):
        """Return the Fortran object for the definition that matches the
        Fortran file, line number, column number

        Parameters
        ----------
        def_file : fortran_file
            File to query
        def_line : int
            Line position in the file
        def_char : int
            Column position in the file
        hover_req : bool, optional
            Flag to enable if calling from a hover request, by default False

        Returns
        -------
        fortran_var | fortran_include | None
            Fortran object
        """
        # Get full line (and possible continuations) from file
        pre_lines, curr_line, _ = def_file.get_code_line(
            def_line, forward=False, strip_comment=True
        )
        # Returns none for string literals, when the query is in the middle
        line_prefix = get_line_prefix(pre_lines, curr_line, def_char, qs=False)
        if line_prefix is None:
            return None
        is_member = False
        try:
            var_stack = get_var_stack(line_prefix)
            is_member = len(var_stack) > 1
            def_name = expand_name(curr_line, def_char)
        except (TypeError, AttributeError):
            return None
        if def_name == "":
            return None
        # Search in Preprocessor defined variables
        if def_name in def_file.pp_defs:
            var = fortran_var(
                def_file.ast,
                def_line + 1,
                def_name,
                f"#define {def_name} {def_file.pp_defs.get(def_name)}",
                [],
            )
            return var
        curr_scope = def_file.ast.get_inner_scope(def_line + 1)
        # Traverse type tree if necessary
        if is_member:
            type_scope = climb_type_tree(var_stack, curr_scope, self.obj_tree)
            # Set enclosing type as scope
            if type_scope is None:
                return None
            else:
                curr_scope = type_scope
        # Find in available scopes
        var_obj = None
        if curr_scope is not None:
            if (
                (curr_scope.get_type() == CLASS_TYPE_ID)
                and (not is_member)
                and (
                    (
                        line_prefix.lstrip().lower().startswith("procedure")
                        and (line_prefix.count("=>") > 0)
                    )
                    or TYPE_DEF_REGEX.match(line_prefix)
                )
            ):
                curr_scope = curr_scope.parent
            var_obj = find_in_scope(
                curr_scope, def_name, self.obj_tree, var_line_number=def_line + 1
            )
        # Search in global scope
        if var_obj is None:
            if is_member:
                return None
            key = def_name.lower()
            if key in self.obj_tree:
                return self.obj_tree[key][0]
            for obj in self.intrinsic_funs:
                if obj.name.lower() == key:
                    return obj

            # If we have a Fortran literal constant e.g. 100, .false., etc.
            # Return a dummy object with the correct type & position in the doc
            if hover_req and curr_scope:
                var_type = None
                if FRegex.NUMBER.match(def_name):
                    if any(s in def_name for s in [".", "e", "d"]):
                        var_type = f"{FORTRAN_LITERAL}REAL"
                    else:
                        var_type = f"{FORTRAN_LITERAL}INTEGER"
                elif FRegex.LOGICAL.match(def_name):
                    var_type = f"{FORTRAN_LITERAL}LOGICAL"
                elif FRegex.SQ_STRING.match(def_name) or FRegex.DQ_STRING.match(
                    def_name
                ):
                    var_type = f"{FORTRAN_LITERAL}STRING"
                if var_type:
                    return fortran_var(
                        curr_scope.file_ast,
                        def_line + 1,
                        def_name,
                        var_type,
                        curr_scope.keywords,
                    )

        else:
            return var_obj
        return None

    def serve_signature(self, request: dict):
        def get_sub_name(line: str):
            _, sections = get_paren_level(line)
            if sections[0].start <= 1:
                return None, None, None
            arg_string = line[sections[0].start : sections[-1].end]
            sub_string, sections = get_paren_level(line[: sections[0].start - 1])
            return sub_string.strip(), arg_string.split(","), sections[-1].start

        def check_optional(arg, params: dict):
            opt_split = arg.split("=")
            if len(opt_split) > 1:
                opt_arg = opt_split[0].strip().lower()
                for i, param in enumerate(params):
                    param_split = param["label"].split("=")[0]
                    if param_split.lower() == opt_arg:
                        return i
            return None

        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        path: str = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        # Check line
        sig_line: int = params["position"]["line"]
        sig_char: int = params["position"]["character"]
        # Get full line (and possible continuations) from file
        pre_lines, curr_line, _ = file_obj.get_code_line(
            sig_line, forward=False, strip_comment=True
        )
        line_prefix = get_line_prefix(pre_lines, curr_line, sig_char)
        if line_prefix is None:
            return None
        # Test if scope declaration or end statement
        if FRegex.SCOPE_DEF.match(curr_line) or FRegex.END.match(curr_line):
            return None
        is_member = False
        try:
            sub_name, arg_strings, sub_end = get_sub_name(line_prefix)
            var_stack = get_var_stack(sub_name)
            is_member = len(var_stack) > 1
        except (TypeError, AttributeError):
            return None
        #
        curr_scope = file_obj.ast.get_inner_scope(sig_line + 1)
        # Traverse type tree if necessary
        if is_member:
            type_scope = climb_type_tree(var_stack, curr_scope, self.obj_tree)
            # Set enclosing type as scope
            if type_scope is None:
                curr_scope = None
            else:
                curr_scope = type_scope
        sub_name = var_stack[-1]
        # Find in available scopes
        var_obj = None
        if curr_scope is not None:
            var_obj = find_in_scope(curr_scope, sub_name, self.obj_tree)
        # Search in global scope
        if var_obj is None:
            key = sub_name.lower()
            if key in self.obj_tree:
                var_obj = self.obj_tree[key][0]
            else:
                for obj in self.intrinsic_funs:
                    if obj.name.lower() == key:
                        var_obj = obj
                        break
        # Check keywords
        if (var_obj is None) and (
            FRegex.INT_STMNT.match(line_prefix[:sub_end]) is not None
        ):
            key = sub_name.lower()
            for candidate in get_intrinsic_keywords(self.statements, self.keywords, 0):
                if candidate.name.lower() == key:
                    var_obj = candidate
                    break
        if var_obj is None:
            return None
        # Build signature
        label, doc_str, params = var_obj.get_signature()
        if label is None:
            return None
        # Find current parameter by index or by
        # looking at last arg with optional name
        param_num = len(arg_strings) - 1
        opt_num = check_optional(arg_strings[-1], params)
        if opt_num is None:
            if len(arg_strings) > 1:
                opt_num = check_optional(arg_strings[-2], params)
                if opt_num is not None:
                    param_num = opt_num + 1
        else:
            param_num = opt_num
        signature = {"label": label, "parameters": params}
        if doc_str is not None:
            signature["documentation"] = doc_str
        req_dict = {"signatures": [signature], "activeParameter": param_num}
        return req_dict

    def get_all_references(
        self,
        def_obj,
        type_mem: bool,
        file_obj: fortran_file = None,
    ):
        # Search through all files
        def_name: str = def_obj.name.lower()
        def_fqsn: str = def_obj.FQSN
        NAME_REGEX = re.compile(rf"(?:\W|^)({def_name})(?:\W|$)", re.I)
        if file_obj is None:
            file_set = self.workspace.items()
        else:
            file_set = ((file_obj.path, file_obj),)
        # A container that includes all the FQSN signatures for objects that
        # are linked to the rename request and that should also be replaced
        override_cache: list[str] = []
        refs = {}
        ref_objs = []
        for filename, file_obj in file_set:
            file_refs = []
            # Search through file line by line
            for (i, line) in enumerate(file_obj.contents_split):
                if len(line) == 0:
                    continue
                # Skip comment lines
                line = file_obj.strip_comment(line)
                if (line == "") or (line[0] == "#"):
                    continue
                for match in NAME_REGEX.finditer(line):
                    var_def = self.get_definition(file_obj, i, match.start(1) + 1)
                    if var_def is None:
                        continue
                    ref_match = False
                    try:
                        # NOTE: throws AttributeError if object is intrinsic since
                        # it will not have a FQSN
                        # BUG: intrinsic objects should be excluded, but get_definition
                        # does not recognise the arguments
                        if def_fqsn == var_def.FQSN or var_def.FQSN in override_cache:
                            ref_match = True
                        # NOTE: throws AttributeError if object is None
                        elif var_def.parent.get_type() == CLASS_TYPE_ID:
                            if type_mem:
                                for inherit_def in var_def.parent.get_overridden(
                                    def_name
                                ):
                                    if def_fqsn == inherit_def.FQSN:
                                        ref_match = True
                                        override_cache.append(var_def.FQSN)
                                        break

                            # Standalone definition of a type-bound procedure,
                            # no pointer replace all its instances in the current scope
                            # NOTE: throws AttributeError if object has no link_obj
                            if (
                                var_def.sline - 1 == i
                                and var_def.file_ast.path == filename
                                and line.count("=>") == 0
                                and var_def.link_obj is def_obj
                            ):
                                ref_objs.append(var_def)
                                override_cache.append(var_def.FQSN)
                                ref_match = True

                        # Object is a Method and the linked object i.e. the
                        # implementation
                        # shares the same parent signature as the current variable
                        # NOTE:: throws and AttributeError if the link_object or
                        # parent are not present OR they are set to None
                        # hence not having a FQSN
                        elif (
                            def_obj.get_type(True) == METH_TYPE_ID
                            and def_obj.link_obj.parent.FQSN == var_def.parent.FQSN
                        ):
                            ref_match = True
                            override_cache.append(var_def.FQSN)
                    except AttributeError:
                        ref_match = False

                    if ref_match:
                        file_refs.append([i, match.start(1), match.end(1)])
            if len(file_refs) > 0:
                refs[filename] = file_refs
        return refs, ref_objs

    def serve_references(self, request):
        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        def_line: int = params["position"]["line"]
        def_char: int = params["position"]["character"]
        path = path_from_uri(uri)
        # Find object
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        def_obj = self.get_definition(file_obj, def_line, def_char)
        if def_obj is None:
            return None
        # Determine global accessibility and type membership
        restrict_file = None
        type_mem = False
        if def_obj.FQSN.count(":") > 2:
            if def_obj.parent.get_type() == CLASS_TYPE_ID:
                type_mem = True
            else:
                restrict_file = def_obj.file_ast.file
                if restrict_file is None:
                    return None
        all_refs, _ = self.get_all_references(def_obj, type_mem, file_obj=restrict_file)
        refs = []
        for (filename, file_refs) in all_refs.items():
            for ref in file_refs:
                refs.append(
                    uri_json(path_to_uri(filename), ref[0], ref[1], ref[0], ref[2])
                )
        return refs

    def serve_definition(self, request: dict):
        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        def_line: int = params["position"]["line"]
        def_char: int = params["position"]["character"]
        path = path_from_uri(uri)
        # Find object
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        var_obj = self.get_definition(file_obj, def_line, def_char)
        if var_obj is None:
            return None
        # Construct link reference
        if var_obj.file_ast.file is not None:
            return self._create_ref_link(var_obj)
        return None

    def serve_hover(self, request: dict):
        def create_hover(string: str, highlight: bool):
            if highlight:
                return {"language": self.hover_language, "value": string}
            else:
                return string

        def create_signature_hover():
            sig_request = request.copy()
            sig_result = self.serve_signature(sig_request)
            try:
                arg_id = sig_result.get("activeParameter")
                if arg_id is not None:
                    arg_info = sig_result["signatures"][0]["parameters"][arg_id]
                    arg_doc = arg_info["documentation"]
                    doc_split = arg_doc.find("\n !!")
                    if doc_split < 0:
                        arg_string = f"{arg_doc} :: {arg_info['label']}"
                    else:
                        arg_string = (
                            f"{arg_doc[:doc_split]} :: "
                            f"{arg_info['label']}{arg_doc[doc_split:]}"
                        )
                    return create_hover(arg_string, True)
            except:
                pass

        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        def_line: int = params["position"]["line"]
        def_char: int = params["position"]["character"]
        path = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        # Find object
        var_obj = self.get_definition(file_obj, def_line, def_char, hover_req=True)
        if var_obj is None:
            return None
        # Construct hover information
        var_type = var_obj.get_type()
        hover_array = []
        if var_type in (SUBROUTINE_TYPE_ID, FUNCTION_TYPE_ID):
            hover_str, highlight = var_obj.get_hover(long=True)
            hover_array.append(create_hover(hover_str, highlight))
        elif var_type == INTERFACE_TYPE_ID:
            for member in var_obj.mems:
                hover_str, highlight = member.get_hover(long=True)
                if hover_str is not None:
                    hover_array.append(create_hover(hover_str, highlight))
        elif var_type == VAR_TYPE_ID:
            # Unless we have a Fortran literal include the desc in the hover msg
            # See get_definition for an explanation about this default name
            if not var_obj.desc.startswith(FORTRAN_LITERAL):
                hover_str, highlight = var_obj.get_hover()
                hover_array.append(create_hover(hover_str, highlight))
            # Hover for Literal variables
            elif var_obj.desc.endswith("REAL"):
                hover_array.append(create_hover("REAL", True))
            elif var_obj.desc.endswith("INTEGER"):
                hover_array.append(create_hover("INTEGER", True))
            elif var_obj.desc.endswith("LOGICAL"):
                hover_array.append(create_hover("LOGICAL", True))
            elif var_obj.desc.endswith("STRING"):
                hover_str = f"CHARACTER(LEN={len(var_obj.name)-2})"
                hover_array.append(create_hover(hover_str, True))

            # Include the signature if one is present e.g. if in an argument list
            if self.hover_signature:
                hover_str = create_signature_hover()
                if hover_str is not None:
                    hover_array.append(hover_str)
        #
        if len(hover_array) > 0:
            return {"contents": hover_array}
        return None

    def serve_implementation(self, request: dict):
        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        def_line: int = params["position"]["line"]
        def_char: int = params["position"]["character"]
        path = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        # Find object
        var_obj = self.get_definition(file_obj, def_line, def_char)
        if var_obj is None:
            return None
        # Intrinsics do not have implementations we can access
        if isinstance(var_obj, fortran_intrinsic_obj):
            return None
        # Construct implementation reference
        if var_obj.parent.get_type() == CLASS_TYPE_ID:
            impl_obj = var_obj.link_obj
            if (impl_obj is not None) and (impl_obj.file_ast.file is not None):
                return self._create_ref_link(impl_obj)
        return None

    def serve_rename(self, request: dict):
        # Get parameters from request
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        def_line: int = params["position"]["line"]
        def_char: int = params["position"]["character"]
        path = path_from_uri(uri)
        # Find object
        file_obj = self.workspace.get(path)
        if file_obj is None:
            return None
        def_obj = self.get_definition(file_obj, def_line, def_char)
        if def_obj is None:
            return None
        if isinstance(def_obj, fortran_intrinsic_obj):
            self.post_message("Rename failed: Cannot rename intrinsics", Severity.warn)
            return None
        # Determine global accesibility and type membership
        restrict_file = None
        type_mem = False
        if def_obj.FQSN.count(":") > 2:
            if def_obj.parent.get_type() == CLASS_TYPE_ID:
                type_mem = True
            else:
                restrict_file = def_obj.file_ast.file
                if restrict_file is None:
                    return None
        all_refs, ref_objs = self.get_all_references(
            def_obj, type_mem, file_obj=restrict_file
        )
        if len(all_refs) == 0:
            self.post_message("Rename failed: No usages found to rename", Severity.warn)
            return None
        # Create rename changes
        new_name = params["newName"]
        changes: dict[str, list[dict]] = {}
        for filename, file_refs in all_refs.items():
            file_uri = path_to_uri(filename)
            changes[file_uri] = []
            for ref in file_refs:
                changes[file_uri].append(
                    change_json(new_name, ref[0], ref[1], ref[0], ref[2])
                )
        return {"changes": changes}

    def serve_codeActions(self, request: dict):
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        sline: int = params["range"]["start"]["line"]
        eline: int = params["range"]["end"]["line"]
        path = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        # Find object
        if file_obj is None:
            return None
        curr_scope = file_obj.ast.get_inner_scope(sline)
        if curr_scope is None:
            return None
        action_list = curr_scope.get_actions(sline, eline)
        if action_list is None:
            return None
        # Convert diagnostics
        for action in action_list:
            diagnostics = action.get("diagnostics")
            if diagnostics is not None:
                new_diags = []
                for diagnostic in diagnostics:
                    new_diags.append(diagnostic.build(file_obj))
                action["diagnostics"] = new_diags
        return action_list

    def send_diagnostics(self, uri: str):
        diag_results, diag_exp = self.get_diagnostics(uri)
        if diag_results is not None:
            self.conn.send_notification(
                "textDocument/publishDiagnostics",
                {"uri": uri, "diagnostics": diag_results},
            )
        elif diag_exp is not None:
            self.conn.write_error(
                -1,
                code=-32603,
                message=str(diag_exp),
                data={
                    "traceback": traceback.format_exc(),
                },
            )

    def get_diagnostics(self, uri: str):
        filepath = path_from_uri(uri)
        file_obj = self.workspace.get(filepath)
        if file_obj is not None:
            try:
                diags = file_obj.check_file(
                    self.obj_tree,
                    max_line_length=self.max_line_length,
                    max_comment_line_length=self.max_comment_line_length,
                )
            except Exception as e:
                return None, e
            else:
                return diags, None
        return None, None

    def serve_onChange(self, request: dict):
        # Update workspace from file sent by editor
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        path = path_from_uri(uri)
        file_obj = self.workspace.get(path)
        if file_obj is None:
            self.post_message(f"Change request failed for unknown file '{path}'")
            return
        else:
            # Update file contents with changes
            reparse_req = True
            if self.sync_type == 1:
                file_obj.apply_change(params["contentChanges"][0])
            else:
                try:
                    reparse_req = False
                    for change in params["contentChanges"]:
                        reparse_flag = file_obj.apply_change(change)
                        reparse_req = reparse_req or reparse_flag
                except:
                    self.post_message(
                        f"Change request failed for file '{path}': Could not apply"
                        " change",
                        Severity.error,
                        exc_info=True,
                    )
                    return
        # Parse newly updated file
        if reparse_req:
            _, err_str = self.update_workspace_file(path, update_links=True)
            if err_str is not None:
                self.post_message(f"Change request failed for file '{path}': {err_str}")
                return
            # Update include statements linking to this file
            for _, tmp_file in self.workspace.items():
                tmp_file.ast.resolve_includes(self.workspace, path=path)
            file_obj.ast.resolve_includes(self.workspace)
            # Update inheritance (currently file only)
            # tmp_file.ast.resolve_links(self.obj_tree, self.link_version)
        elif file_obj.preproc:
            file_obj.preprocess(pp_defs=self.pp_defs)
            self.pp_defs = {**self.pp_defs, **file_obj.pp_defs}

    def serve_onOpen(self, request: dict):
        self.serve_onSave(request, did_open=True)

    def serve_onClose(self, request: dict):
        self.serve_onSave(request, did_close=True)

    def serve_onSave(
        self, request: dict, did_open: bool = False, did_close: bool = False
    ):
        # Update workspace from file on disk
        params: dict = request["params"]
        uri: str = params["textDocument"]["uri"]
        filepath = path_from_uri(uri)
        # Skip update and remove objects if file is deleted
        if did_close and (not os.path.isfile(filepath)):
            # Remove old objects from tree
            file_obj = self.workspace.get(filepath)
            if file_obj is not None:
                ast_old = file_obj.ast
                if ast_old is not None:
                    for key in ast_old.global_dict:
                        self.obj_tree.pop(key, None)
            return
        did_change, err_str = self.update_workspace_file(
            filepath, read_file=True, allow_empty=did_open
        )
        if err_str is not None:
            self.post_message(f"Save request failed for file '{filepath}': {err_str}")
            return
        if did_change:
            # Update include statements linking to this file
            for _, file_obj in self.workspace.items():
                file_obj.ast.resolve_includes(self.workspace, path=filepath)
            file_obj = self.workspace.get(filepath)
            file_obj.ast.resolve_includes(self.workspace)
            # Update inheritance/links
            self.link_version = (self.link_version + 1) % 1000
            for _, file_obj in self.workspace.items():
                file_obj.ast.resolve_links(self.obj_tree, self.link_version)
        if not self.disable_diagnostics:
            self.send_diagnostics(uri)

    def update_workspace_file(
        self,
        filepath: str,
        read_file: bool = False,
        allow_empty: bool = False,
        update_links: bool = False,
    ):
        # Update workspace from file contents and path
        try:
            file_obj = self.workspace.get(filepath)
            if read_file:
                if file_obj is None:
                    file_obj = fortran_file(filepath, self.pp_suffixes)
                    # Create empty file if not yet saved to disk
                    if not os.path.isfile(filepath):
                        if allow_empty:
                            file_obj.ast = fortran_ast(file_obj)
                            self.workspace[filepath] = file_obj
                            return False, None
                        else:
                            return False, "File does not exist"  # Error during load
                err_string, file_changed = file_obj.load_from_disk()
                if err_string:
                    log.error(f"{err_string} : {filepath}")
                    return False, err_string  # Error during file read
                if not file_changed:
                    return False, None
            ast_new = file_obj.parse(
                pp_defs=self.pp_defs, include_dirs=self.include_dirs
            )
            # Add the included read in pp_defs from to the ones specified in the
            # configuration file
            self.pp_defs = {**self.pp_defs, **file_obj.pp_defs}
        except:
            log.error("Error while parsing file %s", filepath, exc_info=True)
            return False, "Error during parsing"  # Error during parsing
        # Remove old objects from tree
        ast_old = file_obj.ast
        if ast_old is not None:
            for key in ast_old.global_dict:
                self.obj_tree.pop(key, None)
        # Add new file to workspace
        file_obj.ast = ast_new
        if filepath not in self.workspace:
            self.workspace[filepath] = file_obj
        # Add top-level objects to object tree
        for key, obj in ast_new.global_dict.items():
            self.obj_tree[key] = [obj, filepath]
        # Update local links/inheritance if necessary
        if update_links:
            self.link_version = (self.link_version + 1) % 1000
            ast_new.resolve_links(self.obj_tree, self.link_version)
        return True, None

    @staticmethod
    def file_init(
        filepath: str,
        pp_defs: dict,
        pp_suffixes: list[str],
        include_dirs: set[str],
        sort: bool,
    ):
        """Initialise a Fortran file

        Parameters
        ----------
        filepath : str
            Path to file
        pp_defs : dict
            Preprocessor definitions
        pp_suffixes : list[str]
            Preprocessor file extension, additional to default
        include_dirs : set[str]
            Preprocessor only include directories, not used by normal parser
        sort : bool
            Whether or not keywords should be sorted

        Returns
        -------
        fortran_file | str
            A Fortran file object or a string containing the error message
        """
        file_obj = fortran_file(filepath, pp_suffixes)
        err_str, _ = file_obj.load_from_disk()
        if err_str:
            return err_str
        try:
            # On Windows multiprocess does not propagate global variables in a shell.
            # Windows uses 'spawn' while Unix uses 'fork' which propagates globals.
            # This is a bypass.
            # For more see on SO: shorturl.at/hwAG1
            set_keyword_ordering(sort)
            file_ast = file_obj.parse(pp_defs=pp_defs, include_dirs=include_dirs)
        except:
            log.error("Error while parsing file %s", filepath, exc_info=True)
            return "Error during parsing"
        file_obj.ast = file_ast
        return file_obj

    def workspace_init(self):
        """Initialise the workspace root across multiple threads"""

        file_list = self._get_source_files()
        # Process files
        pool = Pool(processes=self.nthreads)
        results = {}
        for filepath in file_list:
            results[filepath] = pool.apply_async(
                self.file_init,
                args=(
                    filepath,
                    self.pp_defs,
                    self.pp_suffixes,
                    self.include_dirs,
                    self.sort_keywords,
                ),
            )
        pool.close()
        pool.join()
        for path, result in results.items():
            result_obj = result.get()
            if isinstance(result_obj, str):
                self.post_message(
                    f"Initialization failed for file '{path}': {result_obj}"
                )
                continue
            self.workspace[path] = result_obj
            # Add top-level objects to object tree
            ast_new = self.workspace[path].ast
            for key in ast_new.global_dict:
                self.obj_tree[key] = [ast_new.global_dict[key], path]
        # Update include statements
        for _, file_obj in self.workspace.items():
            file_obj.ast.resolve_includes(self.workspace)
        # Update inheritance/links
        self.link_version = (self.link_version + 1) % 1000
        for _, file_obj in self.workspace.items():
            file_obj.ast.resolve_links(self.obj_tree, self.link_version)

    def serve_exit(self, request: dict) -> None:
        # Exit server
        self.workspace = {}
        self.obj_tree = {}
        self.running = False

    def serve_default(self, request: dict):
        """Raise an error in the Language Server

        Parameters
        ----------
        request : dict
            client dictionary with requests

        Raises
        ------
        JSONRPC2Error
            error with code -32601
        """
        # Default handler (errors!)
        raise JSONRPC2Error(
            code=-32601, message=f"method {request['method']} not found"
        )

    def _load_config_file(self) -> None:
        """Loads the configuration file for the Language Server"""

        # Check for config file
        config_path = os.path.join(self.root_path, self.config)
        if not os.path.isfile(config_path):
            return None

        try:
            with open(config_path, "r") as jsonfile:
                # Allow for jsonc C-style commnets
                # jsondata = "".join(
                #     line for line in jsonfile if not line.startswith("//")
                # )
                # config_dict = json.loads(jsondata)
                config_dict = json.load(jsonfile)

                # Include and Exclude directories
                self._load_config_file_dirs(config_dict)

                # General options
                self._load_config_file_general(config_dict)

                # Preprocessor options
                self._load_config_file_preproc(config_dict)

                # Debug options
                debugging: bool = config_dict.get("debug_log", self.debug_log)
                # If conf option is different than the debug option passed as a
                # command line argument return True so that debug log is setup
                if debugging != self.debug_log and not self.debug_log:
                    self.debug_log = True

        except FileNotFoundError:
            self.post_messages(
                [Severity.error, f"Error settings file '{self.config}' not found"]
            )

        except ValueError:
            self.post_messages(
                [Severity.error, f"Error while parsing '{self.config}' settings file"]
            )

    def _load_config_file_dirs(self, config_dict: dict) -> None:
        # Exclude paths (directories & files)
        # with glob resolution
        for path in config_dict.get("excl_paths", []):
            self.excl_paths.update(set(resolve_globs(path, self.root_path)))

        # Source directory paths (directories)
        # with glob resolution
        source_dirs = config_dict.get("source_dirs", [])
        for path in source_dirs:
            self.source_dirs.update(
                set(
                    only_dirs(
                        resolve_globs(path, self.root_path),
                        self.post_messages,
                    )
                )
            )
        # Keep all directories present in source_dirs but not excl_paths
        self.source_dirs = {i for i in self.source_dirs if i not in self.excl_paths}
        self.incl_suffixes = config_dict.get("incl_suffixes", [])
        # Update the source file REGEX
        self.FORTRAN_SRC_EXT_REGEX = src_file_exts(self.incl_suffixes)
        self.excl_suffixes = set(config_dict.get("excl_suffixes", []))

    def _load_config_file_general(self, config_dict: dict) -> None:
        # General options ------------------------------------------------------
        self.nthreads = config_dict.get("nthreads", self.nthreads)
        self.notify_init = config_dict.get("notify_init", self.notify_init)
        self.incremental_sync = config_dict.get(
            "incremental_sync", self.incremental_sync
        )
        self.sync_type: int = 2 if self.incremental_sync else 1
        self.sort_keywords = config_dict.get("sort_keywords", self.sort_keywords)
        self.disable_autoupdate = config_dict.get(
            "disable_autoupdate", self.disable_autoupdate
        )

        # Autocomplete options -------------------------------------------------
        self.autocomplete_no_prefix = config_dict.get(
            "autocomplete_no_prefix", self.autocomplete_no_prefix
        )
        self.autocomplete_no_snippets = config_dict.get(
            "autocomplete_no_snippets", self.autocomplete_no_snippets
        )
        self.autocomplete_name_only = config_dict.get(
            "autocomplete_name_only", self.autocomplete_name_only
        )
        self.lowercase_intrinsics = config_dict.get(
            "lowercase_intrinsics", self.lowercase_intrinsics
        )
        self.use_signature_help = config_dict.get(
            "use_signature_help", self.use_signature_help
        )

        # Hover options --------------------------------------------------------
        self.hover_signature = config_dict.get("hover_signature", self.hover_signature)
        self.hover_language = config_dict.get("hover_language", self.hover_language)

        # Diagnostic options ---------------------------------------------------
        self.max_line_length = config_dict.get("max_line_length", self.max_line_length)
        self.max_comment_line_length = config_dict.get(
            "max_comment_line_length", self.max_comment_line_length
        )
        self.disable_diagnostics = config_dict.get(
            "disable_diagnostics", self.disable_diagnostics
        )

        # Symbols options ------------------------------------------------------
        self.symbol_skip_mem = config_dict.get("symbol_skip_mem", self.symbol_skip_mem)

        # Code Actions options -------------------------------------------------
        self.enable_code_actions = config_dict.get(
            "enable_code_actions", self.enable_code_actions
        )

    def _load_config_file_preproc(self, config_dict: dict) -> None:
        self.pp_suffixes = config_dict.get("pp_suffixes", None)  # TODO: set def
        self.pp_defs = config_dict.get("pp_defs", {})  # TODO: set other dif?
        if isinstance(self.pp_defs, list):
            self.pp_defs = {key: "" for key in self.pp_defs}

        for path in config_dict.get("include_dirs", set()):
            self.include_dirs.update(
                only_dirs(resolve_globs(path, self.root_path), self.post_messages)
            )

    def _add_source_dirs(self) -> None:
        """Will recursively add all subdirectories that contain Fortran
        source files only if the option `source_dirs` has not been specified
        in the configuration file or no configuration file is present
        """
        # Recursively add sub-directories that only match Fortran extensions
        if len(self.source_dirs) != 1:
            return None
        if self.root_path not in self.source_dirs:
            return None
        self.source_dirs = set()
        for root, dirs, files in os.walk(self.root_path):
            # Match not found
            if not list(filter(self.FORTRAN_SRC_EXT_REGEX.search, files)):
                continue
            if root not in self.source_dirs and root not in self.excl_paths:
                self.source_dirs.add(str(Path(root).resolve()))

    def _get_source_files(self) -> list[str]:
        """Get all the source files present in `self.source_dirs`,
        exclude any files found in `self.excl_paths`^ and ignore
        any files ending with `self.excl_suffixes`.

        ^: the only case where this has not allready happened is when
           `source_dirs` is not specified or a configuration file is not present

        Returns
        -------
        list[str]
            List of source Fortran source files
        """
        # Get filenames
        file_list = []
        for src_dir in self.source_dirs:
            for f in os.listdir(src_dir):
                p = os.path.join(src_dir, f)
                # Process only files
                if not os.path.isfile(p):
                    continue
                # File extension must match supported extensions
                if not self.FORTRAN_SRC_EXT_REGEX.search(f):
                    continue
                # File cannot be in excluded paths/files
                if p in self.excl_paths:
                    continue
                # File cannot have an excluded extension
                if any(f.endswith(ext) for ext in self.excl_suffixes):
                    continue
                file_list.append(p)
        return file_list

    def _config_logger(self, request) -> None:
        """Configures the logger to save Language Server requests/responses to a file
        the logger will by default output to the main (stderr, stdout) channels.
        """

        file_log = True if self.debug_log and self.root_path else False
        fmt = "[%(levelname)-.4s - %(asctime)s] %(message)s"
        if file_log:
            fname = "fortls_debug.log"
            fname = os.path.join(self.root_path, fname)
            logging.basicConfig(filename=fname, level=logging.DEBUG, filemode="w")
            # Also forward logs to the console
            consoleHandler = logging.StreamHandler()
            log.addHandler(consoleHandler)
            log.debug("REQUEST %s %s", request.get("id"), request.get("method"))
            self.post_messages.append([Severity.info, "fortls debugging enabled"])
        else:
            logging.basicConfig(format=fmt, datefmt="%H:%M:%S", level=logging.INFO)

    def _load_intrinsics(self) -> None:
        # Load intrinsics
        set_keyword_ordering(True)  # Always sort intrinsics
        if self.lowercase_intrinsics:
            set_lowercase_intrinsics()
        (
            self.statements,
            self.keywords,
            self.intrinsic_funs,
            self.intrinsic_mods,
        ) = load_intrinsics()
        for module in self.intrinsic_mods:
            self.obj_tree[module.FQSN] = [module, None]
        # Set object settings
        set_keyword_ordering(self.sort_keywords)

    def _create_ref_link(self, obj) -> dict:
        """Create a link reference to an object"""
        obj_file: fortran_file = obj.file_ast.file
        sline, (schar, echar) = obj_file.find_word_in_code_line(obj.sline - 1, obj.name)
        if schar < 0:
            schar = echar = 0
        return uri_json(path_to_uri(obj_file.path), sline, schar, sline, echar)

    def _update_version_pypi(self, test: bool = False):
        """Fetch updates from PyPi for fortls

        Parameters
        ----------
        test : bool, optional
            flag used to override exit checks, only for unittesting, by default False
        """
        if self.disable_autoupdate:
            return False
        # Do not run for prerelease and dev release
        if self._version.is_prerelease and not test:
            return False
        try:
            # For security reasons register as Request before opening
            request = urllib.request.Request("https://pypi.org/pypi/fortls/json")
            with urllib.request.urlopen(request) as resp:
                info = json.loads(resp.read().decode("utf-8"))
                remote_v = version.parse(info["info"]["version"])
                # Do not update from remote if it is a prerelease
                if remote_v.is_prerelease:
                    return False
                # This is the only reliable way to compare version semantics
                if remote_v > self._version or test:
                    self.post_message(
                        "A newer version of fortls is available for download",
                        Severity.info,
                    )
                    # Anaconda environments should handle their updates through conda
                    if os.path.exists(os.path.join(sys.prefix, "conda-meta")):
                        return False
                    self.post_message(
                        f"Downloading from PyPi fortls {info['info']['version']}",
                        Severity.info,
                    )
                    # Run pip
                    result = subprocess.run(
                        [
                            sys.executable,
                            "-m",
                            "pip",
                            "install",
                            "fortls",
                            "--upgrade",
                        ],
                        capture_output=True,
                    )
                    if result.stdout:
                        log.info(result.stdout.decode("utf-8"))
                    if result.stderr:
                        log.error(result.stderr.decode("utf-8"))
                    return True
        # No internet connection exceptions
        except (URLError, KeyError):
            self.post_message("Failed to update the fortls", Severity.warn)
        return False


class JSONRPC2Error(Exception):
    def __init__(self, code, message, data=None):
        self.code = code
        self.message = message
        self.data = data



from setup_tests import run_request, test_dir, write_rpc_request


def hover_req(file_path: str, ln: int, col: int) -> str:
    return write_rpc_request(
        1,
        "textDocument/hover",
        {
            "textDocument": {"uri": str(file_path)},
            "position": {"line": ln, "character": col},
        },
    )


def validate_hover(result_array: list, checks: list):
    assert len(result_array) - 1 == len(checks)
    for (i, check) in enumerate(checks):
        assert result_array[i + 1]["contents"][0]["value"] == check


def test_hover_abstract_int_procedure():
    """Tests that the binding of an abstract interface is correctly resolved"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_abstract.f90"
    string += hover_req(file_path, 7, 30)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """SUBROUTINE test(a, b)
 INTEGER(4), DIMENSION(3,6), INTENT(IN) :: a
 REAL(8), DIMENSION(4), INTENT(OUT) :: b"""
    ]
    validate_hover(results, ref_results)


def test_hover_parameter_multiline():
    """Test that hover parameters display value correctly across lines"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 2, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["INTEGER, PARAMETER :: var = 1000"]
    validate_hover(results, ref_results)


def test_hover_literal_num():
    """Test that hovering over literals shows their type INTEGER"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 3, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["INTEGER"]
    validate_hover(results, ref_results)


def test_hover_parameter():
    """Test that hover parameters display value correctly"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 4, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["INTEGER, PARAMETER :: var2 = 23"]
    validate_hover(results, ref_results)


def test_hover_parameter_nested():
    """Test that hover parameters using other parameter values works"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 4, 41)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["INTEGER, PARAMETER :: var3 = var*var2"]
    validate_hover(results, ref_results)


def test_hover_parameter_multiline_missing_type():
    """Test that hover parameters display correctly when type is split across lines"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 6, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["INTEGER, PARAMETER :: var4 = 123"]
    validate_hover(results, ref_results)


def test_hover_literal_real():
    """Test that hovering over literals shows their values REAL"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 7, 47)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["REAL"]
    validate_hover(results, ref_results)


def test_hover_parameter_double():
    """Test that hovering over parameters shows their type DOUBLE PRECISION"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 7, 38)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["DOUBLE PRECISION, PARAMETER :: somevar = 23.12"]
    validate_hover(results, ref_results)


def test_hover_parameter_double_sf():
    """Test that hovering over parameters shows their type scientific notation"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 7, 55)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["DOUBLE PRECISION, PARAMETER :: some = 1e-19"]
    validate_hover(results, ref_results)


def test_hover_parameter_bool():
    """Test that hovering over parameters shows their values LOGICAL"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 8, 38)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["LOGICAL(kind=8), PARAMETER :: long_bool = .true."]
    validate_hover(results, ref_results)


def test_hover_literal_bool():
    """Test that hovering over literals shows their type LOGICAL"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 8, 50)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["LOGICAL"]
    validate_hover(results, ref_results)


def test_hover_parameter_str_sq():
    """Test that hovering over parameters shows their value, single quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 9, 37)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["CHARACTER(len=5), PARAMETER :: sq_str = '12345'"]
    validate_hover(results, ref_results)


def test_hover_literal_string_sq():
    """Test that hovering over literals shows their values single quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 9, 48)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["CHARACTER(LEN=5)"]
    validate_hover(results, ref_results)


def test_hover_parameter_str_dq():
    """Test that hovering over parameters shows their value, double quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 10, 37)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ['CHARACTER(len=5), PARAMETER :: dq_str = "12345"']
    validate_hover(results, ref_results)


def test_hover_literal_string_dq():
    """Test that hovering over literals shows their values double quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 10, 48)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["CHARACTER(LEN=5)"]
    validate_hover(results, ref_results)


def test_hover_pointer_attr():
    """Test that hovering maintains the variable attributes e.g. POINTER"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "pointers.f90"
    string += hover_req(file_path, 1, 26)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["INTEGER, POINTER"]
    validate_hover(results, ref_results)


def test_hover_functions():
    """Test that hovering over functions provides the expected results"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "functions.f90"
    string += hover_req(file_path, 1, 11)
    string += hover_req(file_path, 7, 19)
    string += hover_req(file_path, 12, 12)
    string += hover_req(file_path, 18, 19)
    string += hover_req(file_path, 23, 34)
    string += hover_req(file_path, 28, 11)
    string += hover_req(file_path, 34, 21)
    string += hover_req(file_path, 46, 11)
    string += hover_req(file_path, 51, 11)
    string += hover_req(file_path, 55, 11)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0

    ref_results = [
        """FUNCTION fun1(arg) RESULT(fun1)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: fun1""",
        """FUNCTION fun2(arg) RESULT(fun2)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: fun2""",
        """FUNCTION fun3(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: retval""",
        """FUNCTION fun4(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: retval""",
        # Notice that the order of the modifiers does not match the source code
        # This is part of the test, ideally they would be identical but previously
        # any modifiers before the type would be discarded
        """PURE ELEMENTAL FUNCTION fun5(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: retval""",
        """FUNCTION fun6(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER, DIMENSION(10,10) :: retval""",
        """PURE FUNCTION outer_product(x, y) RESULT(outer_product)
 REAL, DIMENSION(:), INTENT(IN) :: x
 REAL, DIMENSION(:), INTENT(IN) :: y
 REAL, DIMENSION(SIZE(X), SIZE(Y)) :: outer_product""",
        """FUNCTION dlamch(cmach) RESULT(dlamch)
 CHARACTER :: CMACH""",
        """FUNCTION fun7() RESULT(val)
 TYPE(c_ptr) :: val""",
        """TYPE(c_ptr) FUNCTION c_loc(x) RESULT(c_loc)""",
    ]
    validate_hover(results, ref_results)


def test_hover_spaced_keywords():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "spaced_keywords.f90"
    string += hover_req(file_path, 1, 45)
    string += hover_req(file_path, 2, 99)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """REAL, DIMENSION(:, :), INTENT(IN)""",
        """REAL, DIMENSION( SIZE(ARG1, 1), MAXVAL([SIZE(ARG1, 2), """
        """SIZE(ARG1, 1)]) ), INTENT(OUT)""",
    ]
    validate_hover(results, ref_results)


def test_hover_recursive():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "recursive.f90"
    string += hover_req(file_path, 9, 40)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """RECURSIVE SUBROUTINE recursive_assign_descending(node, vector, current_loc)
 TYPE(tree_inode), POINTER, INTENT(IN) :: node
 INTEGER, DIMENSION(:), INTENT(INOUT) :: vector
 INTEGER, INTENT(INOUT) :: current_loc"""
    ]
    validate_hover(results, ref_results)


def test_hover_subroutine():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_submod.F90"
    string += hover_req(file_path, 29, 24)
    string += hover_req(file_path, 34, 24)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """FUNCTION point_dist(a, b) RESULT(distance)
 TYPE(point), INTENT(IN) :: a
 TYPE(point), INTENT(IN) :: b
 REAL :: distance""",
        """FUNCTION is_point_equal_a(a, b) RESULT(is_point_equal_a)
 TYPE(point), INTENT(IN) :: a
 TYPE(point), INTENT(IN) :: b
 LOGICAL :: is_point_equal_a""",
    ]
    validate_hover(results, ref_results)


def test_hover_interface_as_argument():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_diagnostic_int.f90"
    string += hover_req(file_path, 19, 14)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = (
        # Could be subject to change
        """FUNCTION foo2(f, g, h) RESULT(arg3)
 FUNCTION f(x) :: f
 FUNCTION g(x) :: g
 FUNCTION h(x) :: h
 REAL :: arg3""",
    )
    validate_hover(results, ref_results)



from pathlib import Path

from setup_tests import path_to_uri, run_request, test_dir, write_rpc_request


def validate_refs(result_array, checks):
    def find_in_results(uri, sline):
        for (i, result) in enumerate(result_array):
            if (result["uri"] == uri) and (result["range"]["start"]["line"] == sline):
                del result_array[i]
                return result
        return None

    assert len(result_array) == len(checks)
    for check in checks:
        result = find_in_results(path_to_uri(check[0]), check[1])
        assert result is not None
        assert result["range"]["start"]["character"] == check[2]
        assert result["range"]["end"]["character"] == check[3]


def ref_req(uri: Path, ln: int, ch: int):
    return write_rpc_request(
        2,
        "textDocument/references",
        {
            "textDocument": {"uri": str(uri)},
            "position": {"line": ln - 1, "character": ch - 1},
        },
    )


def test_references():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += ref_req(file_path, 10, 9)
    errcode, results = run_request(string)
    assert errcode == 0
    #
    free_path = str(test_dir / "subdir" / "test_free.f90")
    validate_refs(
        results[1],
        (
            [str(test_dir / "test_prog.f08"), 2, 21, 27],
            [str(test_dir / "test_prog.f08"), 9, 5, 11],
            [free_path, 8, 8, 14],
            [free_path, 16, 9, 15],
            [free_path, 18, 14, 20],
            [free_path, 36, 6, 12],
            [free_path, 44, 6, 12],
            [free_path, 78, 6, 12],
        ),
    )



from pathlib import Path

from setup_tests import run_request, test_dir, write_rpc_request


def sigh_request(uri: Path, line: int, char: int):
    return write_rpc_request(
        1,
        "textDocument/signatureHelp",
        {
            "textDocument": {"uri": str(uri)},
            "position": {"line": line, "character": char},
        },
    )


def validate_sigh(results, refs):
    assert results.get("activeParameter", -1) == refs[0]
    signatures = results.get("signatures")
    assert signatures[0].get("label") == refs[2]
    assert len(signatures[0].get("parameters")) == refs[1]


def test_subroutine_signature_help():
    """Test that the signature help is correctly resolved for all arguments and
    that the autocompletion is correct for the subroutine signature.
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += sigh_request(file_path, 25, 18)
    string += sigh_request(file_path, 25, 20)
    string += sigh_request(file_path, 25, 22)
    string += sigh_request(file_path, 25, 27)
    string += sigh_request(file_path, 25, 29)
    errcode, results = run_request(string)
    assert errcode == 0

    sub_sig = "test_sig_Sub(arg1, arg2, opt1=opt1, opt2=opt2, opt3=opt3)"
    ref = (
        [0, 5, sub_sig],
        [1, 5, sub_sig],
        [2, 5, sub_sig],
        [3, 5, sub_sig],
        [4, 5, sub_sig],
    )
    assert len(ref) == len(results) - 1
    for i, r in enumerate(ref):
        validate_sigh(results[i + 1], r)


def test_intrinsics():

    string = write_rpc_request(
        1, "initialize", {"rootPath": str(test_dir / "signature")}
    )
    file_path = test_dir / "signature" / "nested_sigs.f90"
    string += sigh_request(file_path, 8, 77)
    errcode, results = run_request(
        string, ["--hover_signature", "--use_signature_help", "-n", "1"]
    )
    assert errcode == 0

    ref = [[0, 2, "REAL(A, KIND=kind)"]]
    assert len(ref) == len(results) - 1
    for i, r in enumerate(ref):
        validate_sigh(results[i + 1], r)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import NamedTuple

#: A single line range tuple
Range = NamedTuple("Range", [("start", int), ("end", int)])


@dataclass
class VAR_info:
    """Holds information about a Fortran VARIABLE"""

    var_type: str  #: Type of variable e.g. ``INTEGER``, ``REAL``, etc.
    #: keywords associated with this variable e.g. SAVE, DIMENSION, etc.
    keywords: list[str]  #: Keywords associated with variable
    var_names: list[str]  #: Variable names


@dataclass
class SELECT_info:
    """Holds information about a SELECT construct"""

    type: int  #: Type of SELECT e.g. normal, select type, select kind, select rank
    binding: str  #: Variable/Object being selected upon
    desc: str  #: Description of select e.g. "TYPE", "CLASS", None


@dataclass
class CLASS_info:
    """Holds information about a Fortran CLASS"""

    name: str  #: Class name
    parent: str  #: Parent object of class e.g. ``TYPE, EXTENDS(scaled_vector) :: a``
    keywords: str  #: Keywords associated with the class


@dataclass
class USE_info:
    """Holds information about a Fortran USE statement"""

    mod_name: str  #: Module name
    #: List of procedures, variables, interfaces, etc. imported via only
    only_list: set[str]
    #: A dictionary holding the new names after a rename operation
    rename_map: dict[str, str]


@dataclass
class GEN_info:
    """Holds information about a GENERIC PROCEDURE DEFINITION"""

    bound_name: str  #: Procedure name
    pro_links: list[str]  #: Procedure links
    vis_flag: int  #: Visibility flag, public or private


@dataclass
class SMOD_info:
    """Holds information about Fortran SUBMODULES"""

    name: str  #: Submodule name
    parent: str  #: Submodule i.e. module, parent


@dataclass
class INT_info:
    """Holds information about a Fortran INTERFACE"""

    name: str  #: Interface name
    abstract: bool  #: Whether or not the interface is abstract


@dataclass
class VIS_info:
    """Holds information about the VISIBILITY of a module's contents"""

    type: int  #: Visibility type 0: PUBLIC 1: PRIVATE TODO: convert to boolean
    obj_names: list[str]  #: Module variables, procedures, etc. with that visibility


@dataclass
class INCLUDE_info:
    """Holds information about a Fortran INCLUDE statement"""

    line_number: int  #: Line number of include
    path: str  #: File path to include
    file: None  # fortran_file  #: fortran_file object
    scope_objs: list[str]  #: A list of available scopes


@dataclass
class SUB_info:
    """Holds information about a Fortran SUBROUTINE"""

    name: str  #: Procedure name
    args: str  #: Argument list
    #: Keywords associated with procedure
    keywords: list[str] = field(default_factory=list)
    #: Whether or not this is a ``MODULE PROCEDURE``
    mod_flag: bool = field(default=False)


@dataclass
class RESULT_sig:
    """Holds information about the RESULT section of a Fortran FUNCTION"""

    name: str = field(default=None)  #: Variable name of result
    type: str = field(default=None)  #: Variable type of result
    #: Keywords associated with the result variable, can append without init
    keywords: list[str] = field(default_factory=list)


@dataclass
class FUN_sig(SUB_info):
    """Holds information about a Fortran FUNCTION"""

    #: Function's result with default ``result.name = name``
    result: RESULT_sig = field(default_factory=RESULT_sig)

    def __post_init__(self):
        if not self.result.name:
            self.result.name = self.name



from __future__ import annotations


def range_json(sln: int, sch: int, eln: int = None, ech: int = None):
    return {
        "range": {
            "start": {"line": sln, "character": sch},
            "end": {"line": eln if eln else sln, "character": ech if ech else sch},
        }
    }


def diagnostic_json(sln: int, sch: int, eln: int, ech: int, msg: str, sev: str):
    return {**range_json(sln, sch, eln, ech), "message": msg, "severity": sev}


def uri_json(uri: str, sln: int, sch: int, eln: int = None, ech: int = None):
    return {"uri": uri, **range_json(sln, sch, eln, ech)}


def location_json(uri: str, sln: int, sch: int, eln: int = None, ech: int = None):
    return {"location": uri_json(uri, sln, sch, eln, ech)}


def symbol_json(
    name: str,
    kind: int,
    uri: str,
    sln: int,
    sch: int,
    eln: int = None,
    ech: int = None,
    container_name: str = None,
):
    if container_name:
        return {
            "name": name,
            "kind": kind,
            **location_json(uri, sln, sch, eln, ech),
            "containerName": container_name,
        }
    return {"name": name, "kind": kind, **location_json(uri, sln, sch, eln, ech)}


def change_json(new_text: str, sln: int, sch: int, eln: int = None, ech: int = None):
    return {**range_json(sln, sch, eln, ech), "newText": new_text}



# from types import NoneType
from setup_tests import (
    path_to_uri,
    run_request,
    test_dir,
    write_rpc_notification,
    write_rpc_request,
)


def test_interface_args():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test subroutines and functions with interfaces as arguments
    file_path = str(test_dir / "test_diagnostic_int.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_nonintrinsic():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test that  use, non_intrinsic does not raise a diagnostic error
    file_path = str(test_dir / "test_nonintrinsic.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_submodules_spaced():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test that submodules with spacings in their parent's names are parsed
    file_path = str(test_dir / "test_submodule.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_end_named_variables():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Tests that variables named end do not close the scope prematurely
    file_path = str(test_dir / "diag" / "test_scope_end_name_var.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_external():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test that externals can be split between multiple lines
    # and that diagnostics for multiple definitions of externals can account
    # for that
    file_path = str(test_dir / "diag" / "test_external.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    root = path_to_uri(str((test_dir / "diag" / "test_external.f90").resolve()))
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == [
        {
            "range": {
                "start": {"line": 7, "character": 17},
                "end": {"line": 7, "character": 22},
            },
            "message": 'Variable "VAR_B" declared twice in scope',
            "severity": 1,
            "relatedInformation": [
                {
                    "location": {
                        "uri": str(root),
                        "range": {
                            "start": {"line": 5, "character": 0},
                            "end": {"line": 5, "character": 0},
                        },
                    },
                    "message": "First declaration",
                }
            ],
        },
        {
            "range": {
                "start": {"line": 8, "character": 17},
                "end": {"line": 8, "character": 22},
            },
            "message": 'Variable "VAR_A" declared twice in scope',
            "severity": 1,
            "relatedInformation": [
                {
                    "location": {
                        "uri": str(root),
                        "range": {
                            "start": {"line": 3, "character": 0},
                            "end": {"line": 3, "character": 0},
                        },
                    },
                    "message": "First declaration",
                }
            ],
        },
    ]


def test_forall():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Checks that forall with end forall inside a case select does not cause
    # unexpected end of scope.
    file_path = str(test_dir / "diag" / "test_forall.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_use_ordering():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test USE directive ordering errors
    file_path = str(test_dir / "diag" / "test_use_ordering.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_where():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test where blocks
    file_path = str(test_dir / "diag" / "test_where.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_multiline():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test where semicolon (multi-line)
    file_path = str(test_dir / "diag" / "test_semicolon.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_enum():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test ENUM block
    file_path = str(test_dir / "diag" / "test_enum.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_module_procedure():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test module procedure in submodules importing scopes
    file_path = str(test_dir / "subdir" / "test_submod.F90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == []


def test_max_line_length():
    root = test_dir / "diag"
    string = write_rpc_request(1, "initialize", {"rootPath": str(root)})
    file_path = str(root / "test_lines.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    file_path = str(root / "conf_long_lines.json")
    errcode, results = run_request(string, [f"--config {file_path}"])
    assert errcode == 0
    assert results[1]["diagnostics"] == [
        {
            "range": {
                "start": {"line": 2, "character": 80},
                "end": {"line": 2, "character": 155},
            },
            "message": 'Line length exceeds "max_line_length" (80)',
            "severity": 2,
        },
        {
            "range": {
                "start": {"line": 3, "character": 100},
                "end": {"line": 3, "character": 127},
            },
            "message": 'Comment line length exceeds "max_comment_line_length" (100)',
            "severity": 2,
        },
    ]


def test_implicit_none():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test module procedure in submodules importing scopes
    file_path = str(test_dir / "diag" / "test_implicit_none.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == [
        {
            "range": {
                "start": {"line": 4, "character": 9},
                "end": {"line": 4, "character": 13},
            },
            "message": "IMPLICIT statement without enclosing scope",
            "severity": 1,
        },
    ]


def test_contains():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test module procedure in submodules importing scopes
    file_path = str(test_dir / "diag" / "test_contains.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == [
        {
            "range": {
                "start": {"line": 3, "character": 4},
                "end": {"line": 3, "character": 12},
            },
            "message": "Multiple CONTAINS statements in scope",
            "severity": 1,
        },
        {
            "range": {
                "start": {"line": 5, "character": 0},
                "end": {"line": 5, "character": 8},
            },
            "message": "CONTAINS statement without enclosing scope",
            "severity": 1,
        },
    ]


def test_visibility():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    # Test module procedure in submodules importing scopes
    file_path = str(test_dir / "diag" / "test_visibility.f90")
    string += write_rpc_notification(
        "textDocument/didOpen", {"textDocument": {"uri": file_path}}
    )
    errcode, results = run_request(string)
    assert errcode == 0
    assert results[1]["diagnostics"] == [
        {
            "range": {
                "start": {"line": 3, "character": 0},
                "end": {"line": 3, "character": 0},
            },
            "message": "Visibility statement without enclosing scope",
            "severity": 1,
        }
    ]



from __future__ import annotations

import shlex
import subprocess
import sys
from io import StringIO
from pathlib import Path

root_dir = Path(__file__).parent.parent.resolve()
sys.path.insert(0, root_dir)

# Compromise since isort does not respect noqa
from fortls.jsonrpc import path_to_uri  # noqa: E402, F401
from fortls.jsonrpc import read_rpc_messages  # noqa: E402
from fortls.jsonrpc import write_rpc_notification  # noqa: E402, F401
from fortls.jsonrpc import write_rpc_request  # noqa: E402, F401

test_dir = root_dir / "test" / "test_source"


def check_post_msg(result: dict, msg: str, severity: int):
    assert result["type"] == severity
    assert result["message"] == msg


def run_request(request, fortls_args: list[str] = None):
    command = [
        sys.executable,
        str(root_dir / "fortls.py"),
        "--incremental_sync",
    ]
    if fortls_args:
        # Input args might not be sanitised, fix that
        for i in fortls_args:
            command.extend(shlex.split(i, posix=False))

    pid = subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    results = pid.communicate(input=request.encode())
    tmp_file = StringIO(results[0].decode())
    results = read_rpc_messages(tmp_file)
    parsed_results = []
    for result in results:
        try:
            parsed_results.append(result["result"])
        except KeyError:
            try:
                # Present in `method`s
                parsed_results.append(result["params"])
            except:
                raise RuntimeError(
                    "Only 'result' and 'params' keys have been implemented for testing."
                    " Please add the new key."
                )
        except:
            raise RuntimeError(
                "Unexpected error encountered trying to extract server results"
            )
    errcode = pid.poll()
    return errcode, parsed_results



from setup_tests import run_request, test_dir, write_rpc_request


def validate_comp(result_array, checks):
    assert len(result_array) == checks[0]
    if checks[0] > 0:
        assert result_array[0]["label"] == checks[1]
        assert result_array[0]["detail"] == checks[2]
        try:
            assert result_array[0]["insertText"] == checks[3]
        except KeyError:
            pass


def comp_request(file_path, line, char):
    return write_rpc_request(
        1,
        "textDocument/completion",
        {
            "textDocument": {"uri": str(file_path)},
            "position": {"line": line, "character": char},
        },
    )


def test_comp1():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += comp_request(file_path, 12, 6)
    string += comp_request(file_path, 13, 6)
    string += comp_request(file_path, 17, 24)
    string += comp_request(file_path, 18, 23)
    string += comp_request(file_path, 20, 7)
    string += comp_request(file_path, 21, 20)
    string += comp_request(file_path, 21, 42)
    string += comp_request(file_path, 23, 26)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # test_prog.f08
        [1, "myfun", "DOUBLE PRECISION FUNCTION myfun(n, xval)", "myfun"],
        [9, "glob_sub", "SUBROUTINE glob_sub(n, xval, yval)", "glob_sub"],
        [1, "bound_nopass", "SUBROUTINE bound_nopass(a, b)", "bound_nopass"],
        [1, "bound_pass", "SUBROUTINE bound_pass(arg1)", "bound_pass"],
        [1, "stretch_vector", "TYPE(scaled_vector)"],
        [6, "scale", "TYPE(scale_type)"],
        [2, "n", "INTEGER(4)"],
        [1, "val", "REAL(8)"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp2():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_submod.F90"
    string += comp_request(file_path, 30, 12)
    string += comp_request(file_path, 31, 8)
    string += comp_request(file_path, 31, 23)
    string += comp_request(file_path, 35, 12)
    string += comp_request(file_path, 36, 48)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_submod.F90
        [1, "point", "TYPE"],
        [1, "distance", "REAL"],
        [2, "x", "REAL"],
        [1, "point", "TYPE"],
        [2, "x", "REAL"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp3():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_inc.f90"
    string += comp_request(file_path, 10, 2)
    file_path = test_dir / "subdir" / "test_inc2.f90"
    string += comp_request(file_path, 3, 2)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # test_inc.f90
        [2, "val1", "REAL(8)"],
        # subdir/test_inc2.f90
        [2, "val1", "REAL(8)"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp4():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_abstract.f90"
    string += comp_request(file_path, 7, 12)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_abstract.f90
        [1, "abs_interface", "SUBROUTINE"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp5():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_free.f90"
    string += comp_request(file_path, 10, 22)
    string += comp_request(file_path, 14, 27)
    string += comp_request(file_path, 28, 15)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_free.f90
        [1, "DIMENSION(:)", "KEYWORD"],
        [2, "vector_create", "SUBROUTINE"],
        [3, "INTENT(IN)", "KEYWORD"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp6():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_select.f90"
    string += comp_request(file_path, 21, 7)
    string += comp_request(file_path, 23, 7)
    string += comp_request(file_path, 25, 7)
    string += comp_request(file_path, 30, 7)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_select.f90
        [2, "a", "REAL(8)"],
        [2, "a", "COMPLEX(8)"],
        [1, "n", "INTEGER(4)"],
        [2, "a", "REAL(8)"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp7():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_block.f08"
    string += comp_request(file_path, 2, 2)
    string += comp_request(file_path, 5, 4)
    string += comp_request(file_path, 8, 6)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # test_block.f08
        [9, "READ", "STATEMENT"],
        [10, "READ", "STATEMENT"],
        [11, "READ", "STATEMENT"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp8():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_inherit.f90"
    string += comp_request(file_path, 10, 11)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_inherit.f90
        [1, "val", "REAL(8)"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp9():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_rename.F90"
    string += comp_request(file_path, 13, 5)
    string += comp_request(file_path, 14, 5)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_rename.F90
        [1, "localname", "INTEGER"],
        [2, "renamed_var2", "REAL(8)"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp10():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_vis.f90"
    string += comp_request(file_path, 8, 10)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_vis.f90
        [3, "some_type", "TYPE"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp_import_host_association():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_import.f90"
    string += comp_request(file_path, 15, 20)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # TODO: this should be 1, mytype2 should not appear in autocomplete
        # see #5 and #8 on GitHub
        [2, "mytype", "TYPE"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp_visibility_scopes():
    """Test that PUBLIC, PRIVATE scopes are enforced in autocomplete results."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "completion" / "test_vis_mod_completion.f90"
    string += comp_request(file_path, 12, 16)
    string += comp_request(file_path, 12, 24)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # completion/test_vis_mod_completion.f90
        [1, "some_var", "INTEGER"],
        [3, "length", "INTEGER"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp_interface():
    """Test that the interface signature autocompletion, with placeholders, works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_generic.f90"
    string += comp_request(file_path, 14, 10)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_generic.f90
        [
            4,
            "my_gen",
            "SUBROUTINE my_gen(self, a, b)",
            "my_gen(${1:self}, ${2:a}, ${3:b})",
        ],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp_no_signature_help():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += comp_request(file_path, 12, 6)
    errcode, results = run_request(string)
    assert errcode == 0

    exp_results = (
        # test_prog.f08, completion without signature_help
        # returns the entire completion as a snippet
        [
            1,
            "myfun",
            "DOUBLE PRECISION FUNCTION myfun(n, xval)",
            "myfun(${1:n}, ${2:xval})",
        ],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)


def test_comp_fixed():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_fixed.f"
    string += comp_request(file_path, 15, 8)
    string += comp_request(file_path, 15, 21)
    errcode, results = run_request(string, ["--use_signature_help"])
    assert errcode == 0

    exp_results = (
        # subdir/test_fixed.f90
        [1, "bob", "CHARACTER*(LEN=200)"],
        [1, "dave", "CHARACTER*(20)"],
    )
    assert len(exp_results) == len(results) - 1
    for i, ref in enumerate(exp_results):
        validate_comp(results[i + 1], ref)



import os
import sys
from pathlib import Path

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from fortls.interface import commandline_args  # noqa: E402

parser = commandline_args("fortls")


def test_command_line_general_options():
    args = parser.parse_args(
        "-c config_file.json -n 2 --notify_init --incremental_sync --sort_keywords"
        " --disable_autoupdate --debug_log".split()
    )
    assert args.config == "config_file.json"
    assert args.nthreads == 2
    assert args.notify_init
    assert args.incremental_sync
    assert args.sort_keywords
    assert args.disable_autoupdate
    assert args.debug_log


def test_command_line_file_parsing_options():
    args = parser.parse_args(
        "--source_dirs tmp ./local /usr/include/** --incl_suffixes .FF .fpc .h f20"
        " --excl_suffixes _tmp.f90 _h5hut_tests.F90 --excl_paths exclude tests".split()
    )
    assert args.source_dirs == set(["tmp", "./local", "/usr/include/**"])
    assert args.incl_suffixes == [".FF", ".fpc", ".h", "f20"]
    assert args.excl_suffixes == set(["_tmp.f90", "_h5hut_tests.F90"])
    assert args.excl_paths == set(["exclude", "tests"])


def test_command_line_autocomplete_options():
    args = parser.parse_args(
        "--autocomplete_no_prefix --autocomplete_no_snippets --autocomplete_name_only"
        " --lowercase_intrinsics --use_signature_help".split()
    )
    assert args.autocomplete_no_prefix
    assert args.autocomplete_no_snippets
    assert args.autocomplete_name_only
    assert args.lowercase_intrinsics
    assert args.use_signature_help


def test_command_line_hover_options():
    args = parser.parse_args(
        "--hover_signature --hover_language FortranFreeForm".split()
    )
    assert args.hover_signature
    assert args.hover_language == "FortranFreeForm"


def test_command_line_diagnostic_options():
    args = parser.parse_args(
        "--max_line_length 80 --max_comment_line_length 8 --disable_diagnostics".split()
    )
    assert args.max_line_length == 80
    assert args.max_comment_line_length == 8
    assert args.disable_diagnostics


def test_command_line_preprocessor_options():
    args = parser.parse_args(
        "--pp_suffixes .h .fh --include_dirs /usr/include/** ./local/incl --pp_defs"
        ' {"HAVE_PETSC":"","HAVE_ZOLTAN":"","Mat":"type(tMat)"}'.split()
    )
    assert args.pp_suffixes == [".h", ".fh"]
    assert args.include_dirs == set(["/usr/include/**", "./local/incl"])
    assert args.pp_defs == {"HAVE_PETSC": "", "HAVE_ZOLTAN": "", "Mat": "type(tMat)"}


def test_command_line_symbol_options():
    args = parser.parse_args("--symbol_skip_mem".split())
    assert args.symbol_skip_mem


def test_command_line_code_actions_options():
    args = parser.parse_args("--enable_code_actions".split())
    assert args.enable_code_actions


def unittest_server_init():
    from fortls.langserver import LangServer

    root = (Path(__file__).parent / "test_source").resolve()
    parser = commandline_args("fortls")
    args = parser.parse_args("-c f90_config.json".split())

    server = LangServer(None, vars(args))
    server.root_path = root
    server._load_config_file()

    return server, root


def test_config_file_general_options():
    server, root = unittest_server_init()
    assert server.nthreads == 8
    assert server.notify_init
    assert server.incremental_sync
    assert server.sort_keywords
    assert server.disable_autoupdate


def test_config_file_dir_parsing_options():
    server, r = unittest_server_init()
    # File parsing
    assert server.source_dirs == set(
        [f'{r/"subdir"}', f'{r/"pp"}', f'{r/"pp"/"include"}']
    )
    assert server.incl_suffixes == [".FF", ".fpc", ".h", "f20"]
    assert server.excl_suffixes == set(["_tmp.f90", "_h5hut_tests.F90"])
    assert server.excl_paths == set([f'{r/"excldir"}', f'{r/"hover"}'])


def test_config_file_autocomplete_options():
    server, root = unittest_server_init()
    # Autocomplete options
    assert server.autocomplete_no_prefix
    assert server.autocomplete_no_snippets
    assert server.autocomplete_name_only
    assert server.lowercase_intrinsics
    assert server.use_signature_help


def test_config_file_hover_options():
    server, root = unittest_server_init()
    # Hover options
    assert server.hover_signature
    assert server.hover_language == "FortranFreeForm"


def test_config_file_diagnostic_options():
    server, root = unittest_server_init()
    # Diagnostic options
    assert server.max_line_length == 80
    assert server.max_comment_line_length == 80
    assert server.disable_diagnostics


def test_config_file_preprocessor_options():
    server, root = unittest_server_init()
    # Preprocessor options
    assert server.pp_suffixes == [".h", ".fh"]
    assert server.include_dirs == set([f'{root/"include"}'])
    assert server.pp_defs == {
        "HAVE_PETSC": "",
        "HAVE_ZOLTAN": "",
        "Mat": "type(tMat)",
    }


def test_config_file_symbols_options():
    server, root = unittest_server_init()
    # Symbols options
    assert server.symbol_skip_mem


def test_config_file_codeactions_options():
    server, root = unittest_server_init()
    # Code Actions options
    assert server.enable_code_actions


def test_version_update_pypi():
    from fortls.jsonrpc import JSONRPC2Connection, ReadWriter
    from fortls.langserver import LangServer
    from packaging import version

    parser = commandline_args("fortls")
    args = parser.parse_args("-c f90_config.json".split())
    args = vars(args)
    args["disable_autoupdate"] = False

    stdin, stdout = sys.stdin.buffer, sys.stdout.buffer
    s = LangServer(conn=JSONRPC2Connection(ReadWriter(stdin, stdout)), settings=args)
    s.root_path = (Path(__file__).parent / "test_source").resolve()
    did_update = s._update_version_pypi(test=True)
    assert did_update

    s.disable_autoupdate = True
    did_update = s._update_version_pypi()
    assert not did_update

    s.disable_autoupdate = False
    s._version = version.parse("999.0.0")
    did_update = s._update_version_pypi()
    assert not did_update



from __future__ import annotations

from setup_tests import run_request, test_dir, write_rpc_request


def test_hover():
    def hover_req(file_path: str, ln: int, col: int) -> str:
        return write_rpc_request(
            1,
            "textDocument/hover",
            {
                "textDocument": {"uri": str(file_path)},
                "position": {"line": ln, "character": col},
            },
        )

    def check_return(result_array, checks):
        assert len(result_array) == len(checks)
        for (i, check) in enumerate(checks):
            assert result_array[i]["contents"][0]["value"] == check

    root_dir = test_dir / "pp"
    string = write_rpc_request(1, "initialize", {"rootPath": str(root_dir)})
    file_path = root_dir / "preproc.F90"
    string += hover_req(file_path, 5, 8)  # user defined type
    string += hover_req(file_path, 7, 30)  # variable
    string += hover_req(file_path, 7, 40)  # multi-lin variable
    string += hover_req(file_path, 8, 7)  # function with if conditional
    string += hover_req(file_path, 9, 7)  # multiline function with if conditional
    string += hover_req(file_path, 10, 15)  # defined without ()
    file_path = root_dir / "preproc_keywords.F90"
    string += hover_req(file_path, 6, 2)  # ignores PP across Fortran line continuations
    config = str(root_dir / ".pp_conf.json")
    errcode, results = run_request(string, ["--config", config])
    assert errcode == 0

    # Reference solution
    ref_results = (
        "#define PCType character*(80)",
        "#define PETSC_ERR_INT_OVERFLOW 84",
        "#define varVar 55",
        "#define ewrite if (priority <= 3) write((priority), format)",
        "#define ewrite2 if (priority <= 3) write((priority), format)",
        "#define SUCCESS .true.",
        "REAL, CONTIGUOUS, POINTER, DIMENSION(:)",
    )
    assert len(ref_results) == len(results) - 1
    check_return(results[1:], ref_results)



from pathlib import Path

from setup_tests import path_to_uri, run_request, test_dir, write_rpc_request


def validate_def(result_array, checks):
    # If no definition is given result is None
    if result_array is None:
        assert not checks[0]
        return None
    assert result_array["uri"] == path_to_uri(checks[2])
    assert result_array["range"]["start"]["line"] == checks[0]
    assert result_array["range"]["start"]["line"] == checks[1]


def def_request(uri: Path, line, char):
    return write_rpc_request(
        1,
        "textDocument/definition",
        {
            "textDocument": {"uri": str(uri)},
            "position": {"line": line - 1, "character": char - 1},
        },
    )


def test_def_fun_sub_fixed():
    """Test that going to definition of a function or submodule works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += def_request(file_path, 13, 7)
    string += def_request(file_path, 14, 7)
    errcode, results = run_request(string)
    assert errcode == 0
    fixed_path = str(test_dir / "subdir" / "test_fixed.f")
    ref_res = [[0, 0, fixed_path], [22, 22, fixed_path]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_variable():
    """Test that going to definition of a variable works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += def_request(file_path, 21, 8)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[10, 10, str(test_dir / "test_prog.f08")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_type_bound_procedure1():
    """Test that going to definition of type bound procedure works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += def_request(file_path, 22, 21)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[21, 21, str(test_dir / "subdir" / "test_free.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_type_bound_procedure2():
    """Test that going to definition of type bound procedure works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += def_request(file_path, 22, 43)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[14, 14, str(test_dir / "subdir" / "test_free.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_type_nested_variable():
    """Test that going to definition of type bound nested variables works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_prog.f08"
    string += def_request(file_path, 24, 27)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[5, 5, str(test_dir / "subdir" / "test_free.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_type_in_submod_function():
    """Test that going into the definition of a type bound function in a submodule"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_submod.F90"
    string += def_request(file_path, 31, 13)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[1, 1, str(test_dir / "subdir" / "test_submod.F90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_type_in_submod_procedure():
    """Test that going into the definition of a type bound procedure in a submodule"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_submod.F90"
    string += def_request(file_path, 36, 13)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[1, 1, str(test_dir / "subdir" / "test_submod.F90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_include_file():
    """Test that going into the location of an include file works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_inc.f90"
    string += def_request(file_path, 3, 16)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[2, 2, str(test_dir / "subdir" / "test_inc2.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_include_variable1():
    """Test that going to definition of a variable in an include file works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_inc.f90"
    string += def_request(file_path, 11, 3)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[0, 0, str(test_dir / "subdir" / "test_inc2.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_include_variable2():
    """Test that going to definition of a variable in an include file works."""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_inc2.f90"
    string += def_request(file_path, 4, 3)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[4, 4, str(test_dir / "test_inc.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_include_file_missing():
    """Test that going to the definition of a missing file will not break fortls"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_inc.f90"
    string += def_request(file_path, 13, 14)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[None]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_rename_only_variable():
    """Test that going to definition of a renamed list variable will take you
    to the original definition.
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_rename.F90"
    string += def_request(file_path, 14, 6)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[6, 6, str(test_dir / "subdir" / "test_rename.F90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_rename_only_variable_nested():
    """Test that going to definition of a renamed list variable will take you
    to the original definition, tests the multiply renamed case.
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_rename.F90"
    string += def_request(file_path, 15, 6)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[1, 1, str(test_dir / "subdir" / "test_rename.F90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)


def test_def_function_implicit_result_variable():
    """Test that going to definition on the implicitly defined variable RESULT
    works.
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "functions.f90"
    string += def_request(file_path, 4, 18)
    errcode, results = run_request(string)
    assert errcode == 0
    ref_res = [[3, 3, str(test_dir / "hover" / "functions.f90")]]
    assert len(ref_res) == len(results) - 1
    for i, res in enumerate(ref_res):
        validate_def(results[i + 1], res)



import os

from fparser.common.readfortran import FortranFileReader
from fparser.two.parser import ParserFactory
from fparser.two.utils import Base, walk

os.chdir("/home/gn/Code/Python/fortls/test/test_source")

reader = FortranFileReader("test.f90", include_dirs=["subdir"], ignore_comments=False)
parser = ParserFactory().create(std="f2008")
parse_tree = parser(reader)

for i, node in enumerate(walk(parse_tree)):
    if node is None:
        continue
    if not isinstance(node, Base):
        continue
    print(i, type(node), ": ", node)
